""" Widget: setup tab inside the configuration dialog window """
import logging
from enum import Enum
from pathlib import Path
from typing import Any, Callable
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QFileDialog, QLabel, QMessageBox, QProgressBar, QVBoxLayout, QWidget
from ...fixedStringsJson import exampleDataString, setupText
from ...installationTools import configuration, createShortcut, exampleData
from ...miscTools import hardRestart
from ..guiCommunicate import Communicate
from ..guiStyle import TextButton, widgetAndLayout


class ConfigurationSetup(QWidget):
  """
  Main class
  """
  def __init__(self, comm:Communicate, callbackFinished:Callable[[],None]):
    """
    Initialization

    Args:
      comm (Communicate): communication
      callbackFinished (function): callback function to call upon end
    """
    super().__init__()
    self.comm = comm

    #GUI elements
    self.mainL = QVBoxLayout()
    self.setMinimumWidth(400)
    self.setMinimumHeight(500)
    self.setLayout(self.mainL)
    self.callbackFinished = callbackFinished

    self.mainText = setupText
    self.text = QLabel()
    self.text.setText(self.mainText)
    self.text.setTextFormat(Qt.TextFormat.MarkdownText)
    self.text.setOpenExternalLinks(True)
    self.mainL.addWidget(self.text)
    self.progressBar = QProgressBar(self)
    self.progressBar.setMaximum(24)
    self.progressBar.hide()
    self.mainL.addWidget(self.progressBar)

    _, footerL = widgetAndLayout('H', self.mainL, 's', top='m', left='xl', right='xl')
    style = 'color: orange; font-weight: 550; font-size: 20px'
    self.button1 = TextButton('Click to install / repair', self, [Command.ANALYSE], footerL, style=style)
    self.button2 = TextButton('Click to finish',           self, [Command.FINISHED], footerL, style=style, hide=True)


  def callbackProgress(self, number:int) -> None:
    """
    Increase progressbar by moving to number

    Args:
      number (int): integer to move to
    """
    self.progressBar.setValue(number)
    return


  def execute(self, command:list[Any]) -> None:
    """
    Main method that does all the analysis: open dialogs, ..
    """
    if command[0] is Command.ANALYSE:
      #Configuration
      res = configuration('test', '')
      if res =='':
        self.mainText = self.mainText.replace('- Configuration of preferences','- Configuration of preferences is acceptable' )
        self.text.setText(self.mainText)
      else:
        dirName = QFileDialog.getExistingDirectory(self,'Create and select directory for scientific data',str(Path.home()))
        if dirName is None:
          self.mainText = self.mainText.replace('- Configuration of preferences','- Configuration: user chose to INVALID folder' )
          self.text.setText(self.mainText)
        elif list(Path(dirName).iterdir()):
          button = QMessageBox.question(self, 'PASTA-ELN configuration', 'Folder is not empty. Do you want to remove all content?',
                                  QMessageBox.StandardButton.No, QMessageBox.StandardButton.Yes)
          if button == QMessageBox.StandardButton.Yes:
            configuration('repair', dirName)
          else:
            self.mainText = self.mainText.replace('- Configuration of preferences','- Configuration: user chose to NOT install' )
            self.text.setText(self.mainText)
        else:
          configuration('repair', dirName)
      #Shortcut
      button = QMessageBox.question(self, 'Create shortcut', 'Do you want to create the shortcut for PASTA-ELN on desktop?',
                                    QMessageBox.StandardButton.No, QMessageBox.StandardButton.Yes)
      if button == QMessageBox.StandardButton.Yes:
        createShortcut()
        self.mainText = self.mainText.replace('- Shortcut creation', '- User selected to add a shortcut' )
      else:
        self.mainText = self.mainText.replace('- Shortcut creation', '- User selected to NOT add a shortcut' )
      self.text.setText(self.mainText)
      #Example data
      button = QMessageBox.question(self, 'Example data', exampleDataString,
                                    QMessageBox.StandardButton.No, QMessageBox.StandardButton.Yes)
      if button == QMessageBox.StandardButton.Yes:
        self.progressBar.show()
        exampleData(True, self.callbackProgress)
        self.mainText = self.mainText.replace('- Example data', '- Example data was added')
      else:
        self.mainText = self.mainText.replace('- Example data', '- Example data was NOT added, per user choice')
      self.text.setText(self.mainText)
      #at end
      self.button1.hide()
      self.button2.show()
      logging.info('Setup analyse end')
    elif command[0] is Command.FINISHED:        # What do do when setup is finished: success or unsuccessfully
      hardRestart()
      # self.callbackFinished()
    return


class Command(Enum):
  """ Commands used in this file """
  ANALYSE   = 1
  FINISHED  = 2



""" HTML string parts for HTML export of .eln files """


htmlStart = """
<!DOCTYPE HTML>
<html lang='en'>
<head>
  <meta http-equiv='Content-type' content='text/html;charset=UTF-8' />
  <meta name='viewport' content='width=device-width, initial-scale=1'>
  <meta name='description' content='This is a .eln export from PASTA-ELN'>
  <title>ELN archive HTML preview - PASTA-ELN</title>
  <style>
    body {font-family: 'sans-serif';}
    main {contain: paint;}
    table {
      width: 100%;
      border-collapse: collapse;
      font-family: system-ui, sans-serif;
      margin: 1em 0;
      box-shadow: 0 2px 4px rgba(0, 0, 0, 0.05);
    }
    th, td {
      padding: 0.75rem 1rem;
      text-align: left;
      border-bottom: 1px solid #e2e2e2;
    }
    tbody tr:nth-child(even) {background-color: #c7c7c7;}
    tbody tr:hover {background-color: #afc8ff;}
    th {
      font-weight: 600;
      letter-spacing: 0.5px;
    }
    details {
      padding: 0.5rem;
      border-radius: 0.5rem;
    }
  </style>
</head>
<body>
  <main>
"""

htmlEnd = """
  </main>
  <hr>
  <footer>ELN HTML preview - Generated by PASTA-ELN version ___VERSION___ on ___DATE___</footer>
</body>

</html>
"""

importantKeys = ['name', 'description', 'keywords','hasPart', 'value', 'unitText', 'identifier']



""" Dialog that shows a message and possibly an image """
import qtawesome as qta
from PySide6.QtCore import Qt
from PySide6.QtGui import QTextDocument
from PySide6.QtWidgets import QApplication, QDialog, QLabel, QVBoxLayout, QWidget
from .guiStyle import Image, Label, TextButton, widgetAndLayout

ICON_SIZE = 40                                                             # size of the icon at top of dialog

class MessageDialog(QDialog):
  """ Dialog that shows a message and the progress-bar """
  def __init__(self, parent:QWidget, title:str, text:str, icon:str='', image:str='', minWidth:int=-1) -> None:
    """
    Show message box for little information and possibly an image

    Args:
      parent (QWidget): parent widget (self)
      title (str): title of box
      text (str): text in box
      icon (str): icon: 'Information','Warning','Critical'
      image (str): image to show
      minWidth (int): minimum width of dialog
    """
    super().__init__(parent)
    color = 'red' if icon=='Critical' else '#ffbc00' if icon=='Warning' else '#'
    iconSymbol = qta.icon('fa5s.minus-circle' if icon=='Critical' else
                    'fa5s.exclamation-circle' if icon=='Warning' else
                    'fa5s.info', color='white', scale_factor=1)
    self.setWindowTitle(title)
    if minWidth > 0:
      self.setMinimumWidth(minWidth)
    mainL = QVBoxLayout(self)
    if icon:
      iconLabel = QLabel('')
      iconLabel.setPixmap(iconSymbol.pixmap(ICON_SIZE, ICON_SIZE))
      if color!='#':
        iconLabel.setStyleSheet(f'background: {color};')
      iconLabel.setMinimumSize(ICON_SIZE, ICON_SIZE)
      mainL.addWidget(iconLabel, alignment=Qt.AlignHCenter)                                     # type: ignore
    if image is not None:
      Image(image, mainL, anyDimension=400)
    textLabel = Label(text, 'h2', mainL)
    if text.startswith('<') and text.endswith('>'):
      textLabel.setTextFormat(Qt.TextFormat.RichText)
      text = text.replace('<font color="black">',f'<font color="{parent.comm.palette.get("secondaryText", "").strip()[2:-1]}">')# type: ignore[attr-defined]
      textLabel.setText(text)
    else:
      textLabel.setTextFormat(Qt.TextFormat.MarkdownText)
    # final button box
    _, buttonLineL = widgetAndLayout('H', mainL, 'm', 'm', '0', 'm', 's')
    self.copyButton = TextButton('Copy message', parent, None, buttonLineL, 'Copy to clipboard')
    self.copyButton.clicked.connect(lambda: self.copyToClipboard(text))
    buttonLineL.addStretch(2)
    self.okButton   = TextButton('OK',   parent, None, buttonLineL, 'Accept')
    self.okButton.clicked.connect(self.accept)
    self.setMinimumWidth(800)


  def copyToClipboard(self, text:str) -> None:
    """ Copy text to clipboard
    Args:
      text (str): text to copy
    """
    if text.startswith('<') and text.endswith('>'):
      doc = QTextDocument()
      doc.setHtml(text)
      text = doc.toPlainText()
    clipboard = QApplication.clipboard()
    clipboard.setText(text)


def showMessage(parent:QWidget, title:str, text:str, icon:str='', image:str='', minWidth:int=-1) -> None:
  """
  Show message box for little information and possibly an image

  Args:
    parent (QWidget): parent widget (self)
    title (str): title of box
    text (str): text in box
    icon (str): icon: 'Information','Warning','Critical'
    image (Any): image to show
    minWidth (int): minimum width of dialog
  """
  dialogM = MessageDialog(parent, title, text, icon, image, minWidth)
  dialogM.exec()



"""example add-on: sort direct children of node
"""
from PySide6.QtWidgets import QMessageBox
from pasta_eln.backendWorker.worker import Task
from pasta_eln.miscTools import getHierarchy

# The following two variables are mandatory
description  = 'Sort children'  #short description that is shown in the menu
reqParameter = {} #possibility for required parameters: like API-key, etc. {'API': 'text'}

def main(comm, hierStack, widget, parameter={}):
  """ main function: has to exist and is called by the menu
  Args:
    comm (Communicate): communicate-backend
    hierStack (str): node in hierarchy to start the creation
    widget (QWidget): allows to create new gui dialogs
    parameter (dict): ability to pass parameters

  Returns:
      bool: success
  """
  res = QMessageBox.question(widget, 'Sort direction', 'Sort direct children in ascending order? '
          'If you click no, sort in descending order.',
          buttons=QMessageBox.StandardButtons(QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No | QMessageBox.StandardButton.Cancel),
          defaultButton=QMessageBox.StandardButton.Yes)
  if res == QMessageBox.StandardButton.Cancel:
    return False
  ascendingFlag = res == QMessageBox.StandardButton.Yes
  hierarchy, _ = getHierarchy(comm, hierStack)
  hierStackI = hierStack.split('/')
  for revIdx, node in enumerate(sorted(hierarchy.children, key=lambda n: n.name.lower(), reverse=ascendingFlag)):
    idx = len(hierarchy.children) - 1 - revIdx
    comm.uiRequestTask.emit(Task.MOVE_LEAVES, {'docID':node.id, 'stackOld':hierStackI,'stackNew':hierStackI,
                                               'childOld':node.childNum, 'childNew':idx})
  return True



""" all styling of buttons and other general widgets, some defined colors... """
import logging
from typing import Any, Callable, Optional, Union
import qtawesome as qta
from PySide6.QtCore import QByteArray, Qt
from PySide6.QtGui import QAction, QImage, QKeySequence, QMouseEvent, QPixmap
from PySide6.QtSvgWidgets import QSvgWidget
from PySide6.QtWidgets import (QBoxLayout, QFormLayout, QGridLayout, QHBoxLayout, QLabel, QLayout, QMenu, QMessageBox,
                               QPushButton, QScrollArea, QSizePolicy, QSplitter, QVBoxLayout, QWidget)
from ..textTools.handleDictionaries import dict2ul

space = {'0':0, 's':5, 'm':10, 'l':20, 'xl':80}                                   # spaces: padding and margin


class TextButton(QPushButton):
  """ Button that has only text"""
  def __init__(self, label:str, widget:QWidget, command:list[Any]|None=[], layout:Optional[QLayout]=None,
               tooltip:str='', checkable:bool=False, style:str='', hide:bool=False, iconName:str=''):
    """
    Args:
      label (str): label printed on button
      widget (QWidget): widget / dialog that host the button and that has the execute function
      command (enum): command that is used in called-function: possibly a list of multiple terms
      layout (QLayout): button to be added to this layout
      tooltip (str): tooltip shown when mouse hovers the button
      checkable (bool): can the button change its background color
      style (str): css style
      hide (bool): hidden or shown initially
    """
    super().__init__()
    self.setText(label)
    self.setCheckable(checkable)
    self.setChecked(checkable)
    self.setAutoDefault(False)
    self.setDefault(False)
    if command is not None:
      self.clicked.connect(lambda: widget.execute(command))                       # type: ignore[attr-defined]
    if tooltip:
      self.setToolTip(tooltip)
    if style:
      self.setStyleSheet(style)
    else:
      primaryColor   = widget.comm.palette.get('primary',  'background-color')    # type: ignore[attr-defined]
      secondaryColor = widget.comm.palette.get('buttonText','color')              # type: ignore[attr-defined]
      self.setStyleSheet(f'border-width: 0px; {primaryColor} {secondaryColor}')
    if hide:
      self.hide()
    if iconName:
      icon = qta.icon(iconName, scale_factor=1)
      self.setIcon(icon)
    if layout is not None:
      layout.addWidget(self)


class IconButton(QPushButton):
  """ Button that has only an icon"""
  def __init__(self, iconName:str, widget:QWidget, command:list[Any]=[], layout:Optional[QLayout]=None,
               tooltip:str='', style:str='', hide:bool=False, checkable:bool=False):
    """
    Args:
      iconName (str): icon to show on button
      widget (QWidget): widget / dialog that host the button and that has the execute function
      command (enum): command that is used in called-function: possibly a list of multiple terms
      layout (QLayout): button to be added to this layout
      tooltip (str): tooltip shown when mouse hovers the button
      style (str): css style
      hide (bool): hidden or shown initially
      checkable (bool): can the button change its background color
    """
    super().__init__()
    icon = qta.icon(iconName, scale_factor=1)                                              # color change here
    self.setIcon(icon)
    self.setCheckable(checkable)
    self.command = command
    self.clicked.connect(lambda: widget.execute(self.command))                    # type: ignore[attr-defined]
    self.setFixedHeight(30)
    if tooltip:
      self.setToolTip(tooltip)
    if style:
      self.setStyleSheet(f'QPushButton {{{style}}} QPushButton:checked {{border: "red"; border-width: 5px; {style}}}')
    else:
      primaryColor   = widget.comm.palette.get('primary', 'background-color')     # type: ignore[attr-defined]
      secondaryColor = widget.comm.palette.get('secondary','color')               # type: ignore[attr-defined]
      self.setStyleSheet(f'QPushButton {{border-width: 0px; {primaryColor} {secondaryColor}}} QPushButton:checked {{border: 3px solid red; {primaryColor} {secondaryColor}}}')
    if hide:
      self.hide()
    if layout is not None:
      layout.addWidget(self)


class Action(QAction):
  """ QAction and assign function to menu"""
  def __init__(self, label:str, widget:QWidget, command:list[Any],
               menu:QMenu, shortcut:Optional[str]=None, icon:str=''):
    """
    Args:
      label (str): label printed on submenu
      widget (QWidget): widget / dialog that host the button and that has the execute function
      command (enum): command that is used in called-function: possibly a list of multiple terms
      menu (QMenu): button to be added to this menu
      shortcut (str): shortcut (e.g. Ctrl+K)
      icon (str): icon name
    """
    super().__init__()
    self.setParent(widget)
    self.setText(label)
    self.triggered.connect(lambda : widget.execute(command))                      # type: ignore[attr-defined]
    if icon:
      color = 'black' if widget is None else widget.comm.palette.text             # type: ignore[attr-defined]
      self.setIcon(qta.icon(icon, color=color, scale_factor=1))
    if shortcut is not None:
      self.setShortcut(QKeySequence(shortcut))
    menu.addAction(self)


class Image():
  """ Image widget depending on type of data """
  def __init__(self, data:str, layout:Optional[QLayout], width:int=-1, height:int=-1, anyDimension:int=-1):
    """
    Args:
      data (str): image data in byte64-encoding or svg-encoding
      layout (QLayout): to be added to this layout
      width (int): width of image, dominant if both are given
      height (int): height of image
      anyDimension (int): maximum size in any direction
    """
    if data.startswith('data:image/'):                                                      # jpg or png image
      try:
        byteArr = QByteArray.fromBase64(bytearray(data[22:] if data[21]==',' else data[23:], encoding='utf-8'))
        imageW = QImage()
        imageType = data[11:15].upper()
        success = imageW.loadFromData(byteArr, format=imageType[:-1] if imageType.endswith(';') else imageType)#type: ignore[arg-type]
        if not success:
          logging.warning('Could not load image data with format %s', imageType)
          return
        pixmap = QPixmap.fromImage(imageW)
        if height>0:
          pixmap = pixmap.scaledToHeight(height)
        if width>0:
          pixmap = pixmap.scaledToWidth(width)
        if anyDimension>0:
          if pixmap.size().height()>pixmap.size().width():
            pixmap = pixmap.scaledToHeight(anyDimension)
          else:
            pixmap = pixmap.scaledToWidth(anyDimension)
        label = QLabel()
        label.setPixmap(pixmap)
        label.setAlignment(Qt.AlignCenter)                                                      # type: ignore
        if layout is not None:
          layout.addWidget(label, alignment=Qt.AlignHCenter)                                    # type: ignore
      except Exception as e:
        logging.warning('Error processing base64-image %s', e)
    elif data.startswith('<?xml'):                                                                  #svg image
      imageSVG = QSvgWidget()
      policy = imageSVG.sizePolicy()
      policy.setHorizontalPolicy(QSizePolicy.Policy.Fixed)
      policy.setVerticalPolicy(QSizePolicy.Policy.Fixed)
      imageSVG.setSizePolicy(policy)
      imageSVG.load(bytearray(data, encoding='utf-8'))
      if height>0:
        imageSVG.setMaximumSize(int(float(imageSVG.width())/float(imageSVG.height())*height) ,height)
      if width>0:
        imageSVG.setMaximumSize(width, int(float(imageSVG.height())/float(imageSVG.width())*width))
      if anyDimension>0:
        if imageSVG.height()>imageSVG.width():
          imageSVG.setMaximumSize(int(float(imageSVG.width())/float(imageSVG.height())*anyDimension) ,anyDimension)
        else:
          imageSVG.setMaximumSize(anyDimension, int(float(imageSVG.height())/float(imageSVG.width())*anyDimension))
      if layout is not None:
        layout.addWidget(imageSVG, alignment=Qt.AlignHCenter)                                   # type: ignore
    elif len(data)>2:
      logging.error('guiStyle.Image: %s', data[:50], exc_info=True)
    return


class Label(QLabel):
  """ Label widget: headline, ... """
  def __init__(self, text:str='', size:str='', layout:Optional[QLayout]=None,
               function:Optional[Callable[[str, str],None]]=None, docID:str='', tooltip:str='', style:str=''):
    """ Label widget with given font-size and functions:
    - text selection: if only character, easy selection
      - if formatted text: right-mouse-button to select all (There is no other way, apparently)

    Args:
      text (str): text on label
      size (str): size ['h1','h2','h3']
      layout (QLayout): layout to which to add the label
      function (function): function to call on mouse click
      docID (str): docID on other string to connect to this label
      tooltip (str): tooltip shown when mouse hovers the button
      style (str): css style
    """
    super().__init__()
    self.setText(text)
    if text.startswith('#') or text.startswith('<'):
      self.setTextFormat(Qt.TextFormat.RichText)
    self.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse | Qt.TextInteractionFlag.LinksAccessibleByMouse)
    if size == 'h1':
      style += 'font-size: 14pt'
    elif size == 'h2':
      style += 'font-size: 12pt'
    elif size == 'h3':
      style += 'font-size: 10pt'
    if style:
      self.setStyleSheet(style)
    if layout is not None:
      layout.addWidget(self)
    self.mouseFunction = function
    self.identifier = docID
    if tooltip != '':
      self.setToolTip(tooltip)
    return

  def mousePressEvent(self, _:QMouseEvent) -> None:
    """
    Event after mouse press: only use internal members, not the event itself
    """
    if self.mouseFunction is not None:
      self.mouseFunction(self.text(), self.identifier)
    return


class ScrollMessageBox(QMessageBox):
  """ Scrollable message box for lots of dictionary information """
  def __init__(self, title:str, content:dict[str,Any], style:str=''):
    """
    Args:
      title (str): title
      content (dict): dictionary of lots of information
      style (str): css style
    """
    cssStyle = '<style> ul {list-style-type: none; padding-left: 0; margin: 0; text-indent: -20px; padding-left: -20px;} </style>'
    QMessageBox.__init__(self)
    self.setWindowTitle(title)
    if not style:
      self.setStyleSheet('QScrollArea{min-width:300 px; min-height: 400px}')
    else:
      self.setStyleSheet(style)
    scroll = QScrollArea(self)
    scroll.setWidgetResizable(True)
    self.content = QLabel()
    self.content.setWordWrap(True)
    self.content.setText(cssStyle+dict2ul(content))
    self.content.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)
    scroll.setWidget(self.content)
    self.layout().addWidget(scroll, 0, 0, 1, self.layout().columnCount())                       # type: ignore


def widgetAndLayout(direction:str='V', parentLayout:Optional[Union[QLayout,QSplitter]]=None, spacing:str='0', left:str='0',
                    top:str='0', right:str='0', bottom:str='0') -> tuple[QWidget, QBoxLayout]:
  """
  Convenient function for widget and a boxLayout

  Spacings and margins:
  - different than in css/html
  - spacing is the space between elements in the orientation of the BoxLayout
  - is the padding that surrounds the content in the layout

  Distances are given in
  - '0': zero distance
  - 's': small distance used as padding round elements, or vertical spacings
  - 'm': medium used as space between horizontal elements
  - 'l': large used when things need to be separated
  - 'xl': extra large indentations, frames

  Args:
    direction (str): type of layout [H,V]
    parentLayout (QLayout): to which layout should the widget be added. If none, no adding
    spacing (str): spacing
    left (str): padding on left
    top (str): padding on top
    right (str): padding on right
    bottom (str): padding on bottom
  """
  widget = QWidget()
  layout = QVBoxLayout(widget) if direction=='V' else QHBoxLayout(widget)
  layout.setSpacing(space[spacing])
  layout.setContentsMargins(space[left], space[top], space[right], space[bottom])
  if parentLayout is not None:
    parentLayout.addWidget(widget)
  return widget, layout


def widgetAndLayoutForm(parentLayout:Optional[Union[QLayout,QSplitter]]=None, spacing:str='0', left:str='0',
                    top:str='0', right:str='0', bottom:str='0') -> tuple[QWidget, QFormLayout]:
  """
  Convenient function for widget and a form layout
  - comment see above

  Args:
    parentLayout (QLayout): to which layout should the widget be added. If none, no adding
    spacing (str): spacing
    left (str): padding on left
    top (str): padding on top
    right (str): padding on right
    bottom (str): padding on bottom
  """
  widget = QWidget()
  layout = QFormLayout(widget)
  layout.setSpacing(space[spacing])
  layout.setContentsMargins(space[left], space[top], space[right], space[bottom])
  if parentLayout is not None:
    parentLayout.addWidget(widget)
  return widget, layout


def widgetAndLayoutGrid(parentLayout:Optional[QLayout]=None, spacing:str='0', left:str='0',
                    top:str='0', right:str='0', bottom:str='0') -> tuple[QWidget, QGridLayout]:
  """
  Convenient function for widget and a grid layout
  - comment see above

  Args:
    parentLayout (QLayout): to which layout should the widget be added. If none, no adding
    spacing (str): spacing
    left (str): padding on left
    top (str): padding on top
    right (str): padding on right
    bottom (str): padding on bottom
  """
  widget = QWidget()
  layout = QGridLayout(widget)
  layout.setSpacing(space[spacing])
  layout.setContentsMargins(space[left], space[top], space[right], space[bottom])
  if parentLayout is not None:
    parentLayout.addWidget(widget)
  return widget, layout



"""example add-on: create a report from within the project view

THIS IS A VERY ADVANCED ADD-ON TUTORIAL
This tutorial teaches
- the basic structure of project-view-add-ons (header, function for each node, body, footer)
- this add-on runs as part of the frontend worker (show a GUI element)
  -  to get the data one has to use the signal system to communicate to the backend worker
- the data collection works as part of this system
  - define a variable to store the data
  - define a function to fill the data
  - say that this function should be used whenever new data arrives
  - request new data
  - do a while loop until all data is here: wait
"""
import base64
import re
from io import BytesIO
from pathlib import Path
from anytree import PreOrderIter
from PIL import Image
from PySide6.QtGui import QTextDocument  # This is used for html-markdown conversion: works fine here
from PySide6.QtWidgets import QFileDialog
import pasta_eln
from pasta_eln.miscTools import getDoc, getHierarchy
from pasta_eln.Resources import Icons as icons
from pasta_eln.textTools.stringChanges import markdownEqualizer

# The following two variables are mandatory
description  = 'Create html report'  #short description that is shown in the menu
reqParameter = {} #possibility for required parameters: like API-key, etc. {'API': 'text'}

HTML_HEADER = '<!DOCTYPE html>\n<html>\n<head>\n<style>'\
    '* {font-family: sans-serif;}'\
    '.node {border: 1px dotted #AAAAAA; background-color: #EEEEEE;}'\
    'h1,h2,h3,h4 {margin-bottom: 0;} p {margin-top: 0;}'\
    'table {width: 100%;} td {width: 50%; vertical-align: top;}'\
    '.footline {margin-top: 50px; text-align: right;}'\
    '</style>\n</head>\n<body>\n'
HTML_FOOTER = '</body>\n</html>\n'

def main(comm, hierStack, widget, parameter={}):
    """ main function: has to exist and is called by the menu
    Args:
        comm (Communicate): communicate-backend
        hierStack (str): node in hierarchy to start the creation
        widget (QWidget): allows to create new gui dialogs
        parameter (dict): ability to pass parameters

    Returns:
        bool: success
    """
    # get filename to write into
    if 'fileNames' not in parameter:
        res = QFileDialog.getSaveFileName(widget,'Use this file for output', str(Path.home()))
        if res is None:
            return False
    else:
        res = parameter['fileNames']
    hierarchy, _ = getHierarchy(comm, hierStack.split('/')[0])
    qtDocument = QTextDocument()   #used for markdown -> html conversion

    # function to handle each data entry
    def node2html(node):
        """
        Function that renders each node into html
        - Node properties: depth, name, docType, id

        Args:
            node (anyNode): anytree node to process

        Returns:
            str: conversion of node into html string
        """
        hidden = not all(node.gui)        # is this node hidden?
        if hidden:
            return ''
        doc = getDoc(comm, node.id)  # GET ALL INFORMATION OF THIS NODE
        output = '<div class="node">\n'
        # headline of each node: either as html headline or normal text, incl. the objective
        if node.depth<4 and node.docType[0][0]=='x':
            output += f'<h{node.depth+1}>{node.name}</h{node.depth+1}>'
        else:
            output += f"{node.name}&nbsp;&nbsp;{'/'.join(node.docType)}<br>\n"
        if node.docType[0]=='x0':
            output += f"<b>Objective: {doc.get('',{}).get('objective',[''])[0]}</b>"
        # create a two column table: left side: tags and comments; right side: image, content
        output += '<table><tr><td>'
        stars = ''.join('\u2605'*int(i[1]) for i in doc['tags'] if re.match(r'^_\d$',i))  # the stars are part of tags
        otherTags = ', '.join(i for i in doc['tags'] if not re.match(r'^_\d$',i))
        output += f'Tags: {stars} {otherTags}<br>' if stars or otherTags else ''
        qtDocument.setMarkdown(markdownEqualizer(doc['comment']))
        output += qtDocument.toHtml()
        output += '</td><td>'
        if 'content' in doc and doc['content']:
            qtDocument.setMarkdown(markdownEqualizer(doc['content']))
            output += qtDocument.toHtml()
        output += f'<img src="{doc["image"]}"/>' if 'image' in doc and doc['image'].startswith('data:') else ''
        output += doc['image'] if 'image' in doc and doc['image'].startswith('<?xml version="1.0"') else ''
        return f'{output}</td></tr></table></div>\n\n'

    # main function that calls the render function
    out = ''.join(node2html(node) for node in PreOrderIter(hierarchy))

    # add footer line with pasta-eln icon (read and converted to base64 to be inline included in html)
    iconImg = Image.open(f'{icons.__path__[0]}/favicon64.ico')
    figfile = BytesIO()
    iconImg.save(figfile, format='PNG')
    imageB64 = base64.b64encode(figfile.getvalue()).decode()
    imageB64 = f"data:image/png;base64,{imageB64}"
    out += f'<div class="footline">Created with Pasta-ELN {pasta_eln.__version__} and the default HTML export&nbsp;&nbsp;&nbsp;<image src="{imageB64}"/></div>'

    # save everything to the html file
    with open(res[0], 'w', encoding='utf-8') as f:
        f.write(f'{HTML_HEADER}{out}{HTML_FOOTER}')
    return True



""" MandatoryColumnDelegate  module used for the table views """
#  PASTA-ELN and all its sub-parts are covered by the MIT license
#
#  Copyright (c) 2023
#
#  Author: Jithu Murugan
#  Filename: mandatory_column_delegate.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information

from typing import Union
from PySide6.QtCore import QAbstractItemModel, QEvent, QModelIndex, QPersistentModelIndex, QRect
from PySide6.QtGui import QPainter
from PySide6.QtWidgets import (QApplication, QRadioButton, QStyle, QStyledItemDelegate, QStyleOptionButton,
                               QStyleOptionViewItem, QWidget)


class MandatoryColumnDelegate(QStyledItemDelegate):
  """
  Delegate for creating the radio buttons for the mandatory column in data hierarchy editor tables
  """
  def paint(self,
            painter: QPainter,
            option: QStyleOptionViewItem,
            index: Union[QModelIndex, QPersistentModelIndex]) -> None:
    """
    Draws the mandatory radio button within the cell represented by index
    Args:
      painter (QPainter): Painter instance for painting the button
      option (QStyleOptionViewItem): Style option for the cell represented by index
      index (Union[QModelIndex, QPersistentModelIndex]): Table cell index

    Returns: None

    """
    indexName = index.model().index(index.row(), 0)
    if not indexName.data():
      return
    widget = option.widget                                                        # type: ignore[attr-defined]
    style = widget.style() if widget else QApplication.style()
    opt = QStyleOptionButton()
    radio_button = QRadioButton()
    option_rect = option.rect                                                     # type: ignore[attr-defined]
    opt.rect = QRect(option_rect.left() + option_rect.width() / 2 - 10,           # type: ignore[attr-defined]
                     option_rect.top(),
                     option_rect.width(),
                     option_rect.height())
    indexName = index.model().index(index.row(), 0)
    isMandatory = index.data()=='T' and indexName.data()
    opt.state = QStyle.StateFlag.State_On if isMandatory else QStyle.StateFlag.State_Off# type: ignore[attr-defined]
    style.drawControl(QStyle.ControlElement.CE_RadioButton, opt, painter, radio_button)


  def editorEvent(self,
                  event: QEvent,
                  model: QAbstractItemModel,
                  option: QStyleOptionViewItem,
                  index: Union[QModelIndex, QPersistentModelIndex]) -> bool:
    """
    In case of mouse click event, the model data is toggled for the respective table cell index
    Args:
      event (QEvent): The editor event information
      model (QAbstractItemModel): Model data representing the table view
      option (QStyleOptionViewItem): QStyleOption for the table cell
      index (Union[QModelIndex, QPersistentModelIndex]): Table cell index

    Returns (bool): True/False
    """
    if event.type() == QEvent.Type.MouseButtonRelease:
      indexName = model.index(index.row(), 0)
      if indexName.data():
        isMandatory = index.data()=='T'
        model.setData(index, '' if isMandatory else 'T')                        # invert mandatory after click
    return super().editorEvent(event, model, option, index)


  def createEditor(self,
                   parent: QWidget,
                   option: QStyleOptionViewItem,
                   index: Union[QModelIndex, QPersistentModelIndex]) -> QWidget:
    """
    Disable the editor for the whole mandatory column by simply returning None
    Args:
      parent (QWidget): Parent table view
      option (QStyleOptionViewItem): Style option for the cell represented by index
      index (Union[QModelIndex, QPersistentModelIndex]): Cell index

    Returns: None
    """
    return None                                                                   # type: ignore[return-value]



#!/usr/bin/python3
"""TEST using the FULL set of python-requirements: create the default example that all installations create and verify it thoroughly """
import logging, warnings, unittest, tempfile, os
from pathlib import Path
from zipfile import ZipFile
from pasta_eln.backendWorker.backend import Backend
from pasta_eln.backendWorker.inputOutput import exportELN
from pasta_eln.miscTools import getConfiguration


class TestStringMethods(unittest.TestCase):
  """
  derived class for this test
  """


  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.be = None
    self.dirName = ''


  def test_main(self):
    """
    main function
    """
    # initialization: create database, destroy on filesystem and database and then create new one
    warnings.filterwarnings('ignore', message='numpy.ufunc size changed')
    warnings.filterwarnings('ignore', message='invalid escape sequence')
    warnings.filterwarnings('ignore', category=ResourceWarning, module='PIL')
    warnings.filterwarnings('ignore', category=ImportWarning)
    logPath = Path.home()/'pastaELN.log'
    logging.basicConfig(filename=logPath, level=logging.INFO, format='%(asctime)s|%(levelname)s:%(message)s',
                        datefmt='%m-%d %H:%M:%S')   #This logging is always info, since for installation only
    for package in ['urllib3', 'requests', 'asyncio', 'PIL', 'matplotlib.font_manager']:
      logging.getLogger(package).setLevel(logging.WARNING)
    logging.info('Start test')

    log_records = []
    class ErrorHandler(logging.Handler):
      def emit(self, record):
        if record.levelno >= logging.ERROR:
          log_records.append(record)
    handler = ErrorHandler()
    logging.getLogger().addHandler(handler)

    # create .
    configuration, _ = getConfiguration('research')
    self.be = Backend(configuration, 'research')
    projID = self.be.output('x0').split('\n')
    projID = [x for x in projID if 'PASTAs Example Project' in x]
    projID = projID[0].split('|')[-2].strip()
    tempDir = tempfile.gettempdir()
    fileName = f'{tempDir}/PASTA_ELN.eln'
    allDocTypes = self.be.db.dataHierarchy('','')+['-']
    print(f'Filename {fileName}: docTypes: {", ".join(allDocTypes)}')
    exportELN(self.be, [projID], fileName, allDocTypes)

    # test file
    with ZipFile(fileName, 'r') as zObject:
      zObject.extractall(path=tempDir)
    os.system(f'tree {tempDir}/PASTA')
    print("""
Data on disk:
├── pastaELN.db
├── PastasExampleProject
│   ├── 000_ThisIsAnExampleTask
│   ├── 001_ThisIsAnotherExampleTask
│   │   ├── 000_ThisIsAnExampleSubtask
│   │   ├── 001_ThisIsAnotherExampleSubtask
│   │   └── simple.png
│   └── 002_DataFiles
│       ├── simple.csv
│       ├── simple.png
│       └── story.odt
└── CommonFiles
    └── Example_SOP.md
""")

    logging.getLogger().removeHandler(handler)
    self.assertEqual(len(log_records), 0, f"Logging errors found: {[r.getMessage() for r in log_records]}")
    # try:
    #   os.system(f'code {tempDir}/PASTA/ro-crate-metadata.json')
    # except Exception:
    #   pass
    return


  def tearDown(self):
    logging.info('End test')
    return


if __name__ == '__main__':
  unittest.main()



""" Client for communicating with Dataverse Server via REST API
- Author: Jithu Murugan, Steffen Brinckmann
"""
import logging
from datetime import datetime
from json import dumps
from os.path import basename
from typing import Any
from xml.etree.ElementTree import ElementTree, fromstring
# Dataverse: remember to always publish everything before using it!!
import requests
from requests.auth import HTTPBasicAuth
from .dataverseDefaultDict import DATAVERSE_METADATA
from .repository import RepositoryClient


class DataverseClient(RepositoryClient):
  """ Client for communicating with Dataverse Server via REST API """

  def __init__(self, server_url: str, api_token: str, identifier:str) -> None:
    """
    Initializes the client

    Args:
        server_url (str): The URL of the server
        api_token (str): The API token for authentication
        identifier (str): sub-dataverse, category; use '' for void
    """
    super().__init__(server_url, api_token)
    self.identifier = identifier
    self.headers = {'Accept': 'application/json', 'X-Dataverse-key': self.api_token}


  def recreateAPIKey(self) -> str | Any:
    """
    Recreates the API token using existing token

    Returns:
        str | Any: The new API token or any error message if the token recreation fails
    """
    resp = requests.post(f"{self.server_url}/api/users/token/recreate", headers=self.headers, timeout=10)
    if resp.status_code == 200:
      token_message = resp.json().get('data').get('message')
      return token_message.replace('New token for dataverseAdmin is ', '')
    return f"Error recreating the token, Info: {resp.text}"


  def checkServer(self) -> tuple[bool, str]:
    """
    Checks if the data-verse server is reachable

    Returns (tuple(bool, Any)):
      A tuple of (success, a message) is returned
    """
    resp = requests.get(f"{self.server_url}/api/info/version", headers={'Accept': 'application/json'}, timeout=10)
    success = (resp.status_code == 200 and resp.json().get('data').get('version') is not None)
    return (success, 'Dataverse is reachable') \
      if success \
      else (success, f"Cannot reach server: {self.server_url}, Status: {resp.status_code}, json: {resp.json()}")


  def checkAPIKey(self) -> bool:
    """
    Checks if the given API token is valid

    Explanation:
        This method checks if the provided API token is valid by making a request to the server
        It logs the server URL and sends a GET request to the token endpoint with the API token
        It returns True if the response is successful and the status code is not 401, 403, or 500

    Args:
        self: The instance of the class

    Returns:
        bool: True if the API token is valid, False otherwise
    """
    resp = requests.get(f"{self.server_url}/api/users/token", headers=self.headers, timeout=10)
    return (bool(resp) and resp.status_code is not None
            and resp.status_code not in [401, 403, 500])


  def checkAPIKeyExpired(self) -> bool:
    """
    Check if the API token has expired
    Returns:
      True if the token has expired, False otherwise
    """
    resp = requests.get(f"{self.server_url}/api/users/token", headers=self.headers, timeout=10)
    if resp.status_code == 200:
      expiry_message = resp.json().get('data').get('message')
      expiry_time_string = expiry_message.replace(f"Token {self.api_token} expires on ", '')
      expiry_time = datetime.strptime(expiry_time_string, '%Y-%m-%d %H:%M:%S.%f')
      return expiry_time < datetime.now()
    print(f"Error checking token expiration, Info: {resp.text}")
    return False


  ################ DATAVERSE #######################
  def createDataverse(self, dv_parent: str, dv_name: str, dv_description: str, dv_alias: str,
                      dv_contact_email_list: list[dict[str, str]], dv_affiliation: str, dv_type: str) -> dict[Any, Any] | Any:
    """
    Creates and publishes a dataverse
    Args:
      dv_parent (str): The name of the parent dataverse
      dv_name (str): The name of the dataverse
      dv_description (str): The description of the dataverse
      dv_alias (str): The alias of the dataverse
      dv_contact_email_list (list[dict[str, str]]): The list of contact emails which must be provided in the following format:
        [{"contactEmail": "example1@example.edu"}, {"contactEmail": "example2@example.edu"}]
      dv_affiliation (str): The affiliation of the dataverse
      dv_type (str): The type of the dataverse which must be one of the following:
        DEPARTMENT
        JOURNALS
        LABORATORY
        ORGANIZATIONS_INSTITUTIONS
        RESEARCHERS
        RESEARCH_GROUP
        RESEARCH_PROJECTS
        TEACHING_COURSES
        UNCATEGORIZED

    Returns(dict[Any, Any] | Any):
      A dictionary representing the created and published dataverse, or an error message
    """
    dv_json = {
      'name': dv_name,
      'alias': dv_alias,
      'dataverseContacts': dv_contact_email_list,
      'affiliation': dv_affiliation,
      'description': dv_description,
      'dataverseType': dv_type
    }
    # Create the data-verse
    resp = requests.post(f"{self.server_url}/api/dataverses/{dv_parent}",
      headers={'Content-Type': 'application/json', 'X-Dataverse-key': self.api_token},
      data=dumps(dv_json), timeout=10)
    if resp.status_code == 201:                                                                      # Success
      pub_resp = requests.post(f"{self.server_url}/api/dataverses/{resp.json().get('data').get('alias')}/actions/:publish",
        headers={'Content-Type': 'application/json', 'X-Dataverse-key': self.api_token}, timeout=10)
      if pub_resp.status_code == 200:
        return pub_resp.json().get('data')
      return f"Error publishing dataverse, Status: {pub_resp.status_code}, Info: {pub_resp.text}"
    return f"Error creating dataverse, Status: {resp.status_code}, Info: {resp.text}"


  def getDataverseList(self) -> dict[Any, Any] | Any:
    """
    Gets the list of data verses
    Returns:
      A dictionary of dataverses (identifier & title) for successful request,
      otherwise the error message is returned
    """
    resp = requests.get(f"{self.server_url}/dvn/api/data-deposit/v1.1/swordv2/service-document",
      headers={'Accept': 'application/json', 'X-Dataverse-key': self.api_token},
      auth=HTTPBasicAuth(self.api_token, ''), timeout=10)
    if resp.status_code == 200:
      element_tree: ElementTree = ElementTree(fromstring(resp.text))
      root = element_tree.getroot()
      if root is not None:
        dataverse_list: list[dict[str, str]] = []
        for element in root.findall('.//{http://www.w3.org/2007/app}collection'):
          title = element.find('.//{http://www.w3.org/2005/Atom}title')
          if title is not None:
            title_val = title.text if title.text is not None else ''
            dataverse_list.append({'id': element.attrib['href'].split('/')[-1],'title': title_val})
        dataverse_list.sort(key=lambda x: x['title'])
        return dataverse_list
    return f"Error get dataverse list, Server:{self.server_url},  Status:{resp.status_code}"


  def getDataverseContent(self) -> dict[Any, Any] | Any:
    """
    Retrieves the contents of a dataverse

    Returns (dict[Any, Any] | Any):
      A dictionary of dataverse contents for successful request, otherwise the error message is returned
    """
    resp = requests.get(f"{self.server_url}/api/dataverses/{self.identifier}/contents", headers=self.headers,
                        timeout=10)
    return resp.json() if resp.status_code == 200 else \
      f"Error retrieving the contents of dataverse, Id: {self.identifier}, Info: {resp.json()}"


  def getDataverseSize(self) -> str | Any:
    """
    Retrieves the size of a dataverse
    Returns (str):
      Dataverse size in bytes for successful request, otherwise the error message is returned
    """
    resp = requests.get(f"{self.server_url}/api/dataverses/{self.identifier}/storagesize", headers=self.headers,
                        timeout=10)
    if resp.status_code == 200:
      return (resp.json().get('data').get('message').replace('Total size of the files stored in this dataverse: ', ''))
    return f"Error retrieving the size for data verse, Id: {self.identifier}, Info: {resp.json()}"


  def getDataverseInfo(self) -> str | Any:
    """
    Retrieves information about a dataverse

    Returns:
        str | Any: The data associated with the dataverse if the request is successful, otherwise an error message
    """
    resp = requests.get(f"{self.server_url}/api/dataverses/{self.identifier}", headers=self.headers,
                        timeout=10)
    if resp.status_code:
      return resp.json().get('data')
    return f"Error retrieving the info for data verse, Id: {self.identifier}, Info: {resp.text}"


  ################ DATASET #######################
  def createDataset(self, ds_metadata: dict[str, Any], ds_validate_metadata: bool = False
                                       ) -> dict[Any, Any] | Any:
    """
    Creates and publishes a dataset to the parent dataverse
    Args:
      ds_metadata (dict[str, Any]): The dataset metadata
        Refer the https://guides.dataverse.org/en/latest/_downloads/4e04c8120d51efab20e480c6427f139c/dataset-create-new-all-default-fields.json for the default values to be used in the metadata
        The type names to be used in the metadata along with the values can be found in the metadata blocks and should correspond to the dataset-create-new-all-default-fields.json
      ds_validate_metadata (bool): Whether to validate the metadata

    Returns:
      A dictionary of dataset metadata with the persistent identifier for successful request, otherwise the error message is returned
    """
    metadata = DATAVERSE_METADATA
    if 'license' in ds_metadata:
      metadata['datasetVersion']['license'] = ds_metadata['license']
    else:
      del metadata['datasetVersion']['license']
    for _, metablock in metadata['datasetVersion']['metadataBlocks'].items():      # type:ignore[attr-defined]
      field_copy = metablock['fields'].copy()
      del metablock['displayName']
      metablock['fields'].clear()
      for field in field_copy:
        if field['typeName'] in ds_metadata:
          field['value'] = ds_metadata[field['typeName']]
          metablock['fields'].append(field)
    # Request to create the dataset
    resp = requests.post(f"{self.server_url}/api/dataverses/{self.identifier}/datasets",
      params={'doNotValidate': str(not ds_validate_metadata)}, timeout=10,
      headers={'Content-Type': 'application/json', 'X-Dataverse-key': self.api_token}, json=metadata)
    if resp.status_code == 201:
      # Request to publish the dataset
      resp = requests.post(f"{self.server_url}/api/datasets/:persistentId/actions/:publish",
        params={'persistentId': resp.json().get('data').get('persistentId'), 'type': 'major'},
        headers={'Content-Type': 'application/json', 'X-Dataverse-key': self.api_token}, timeout=10)
      if resp.status_code == 200:
        return resp.json().get('data')
      return f"Error publishing dataset, Info: {resp.text}"
    return f"Error creating dataset, Info: {resp.text}"


  def uploadFile(self, ds_pid: str, df_file_path: str, df_description: str, df_categories: list[str]) -> dict[Any, Any] | Any:
    """
    Uploads a file to a dataset
    Args:
      ds_pid (str): The identifier of the dataset
      df_file_path (str): The absolute path to the file to be uploaded
      df_description (str): The description of the file
      df_categories (list[str]): The categories/tags for the file

    Returns:
      {   'file_upload_result': file_upload_response,
          'dataset_publish_result': dataset_publish_response
      } for successful request, otherwise the error message is returned
    """
    filename = basename(df_file_path)
    metadata = dumps({'description': df_description, 'categories': df_categories})
    data:Any = {}
    with open(df_file_path, 'rb') as file_stream:
      data['file'] = (filename, file_stream, 'multipart/form-data')
      data['jsonData'] = (None, metadata, 'application/json')
      # Request to add the file to dataset
      resp = requests.post(
        f"{self.server_url}/api/datasets/:persistentId/add",
        params={'persistentId': ds_pid},
        headers={'X-Dataverse-key': self.api_token},
        files=data,
        timeout=5)
      if resp.status_code == 200:
        # Request to publish the dataset
        pub_resp = requests.post(
          f"{self.server_url}/api/datasets/:persistentId/actions/:publish",
          params={'persistentId': ds_pid, 'type': 'major'}, timeout=10,
          headers={'Content-Type': 'application/json', 'X-Dataverse-key': self.api_token})
        if pub_resp.status_code == 200:
          return {'file_upload_result': resp.json().get('data'),
                  'dataset_publish_result': pub_resp.json().get('data')}
        return f"Error publishing dataset: {ds_pid} as part of file ({df_file_path}) upload on server: "\
               f"{self.server_url}, Info: {pub_resp.json()}"
      return f"Error uploading file: {df_file_path} to dataset: {ds_pid} Info: {resp.json()}"


  def uploadRepository(self, metadata:dict[str,Any], file_path:str) -> tuple[bool, str]:
    """
    Uploads a file and metadata to become a dataset

    Args:
      metadata (dict): metadata to this file according to dataverse standard
      file_path (str): The absolute path to the file to be uploaded

    Returns:
      tuple: success of function, message
    """
    res= self.createDataset(metadata)
    if isinstance(res, str):
      return False, f'Error publishing the dataset: {res}'
    doi = f"{res['protocol']}:{res['authority']}/{res['identifier']}"
    reply = self.uploadFile(doi, file_path, '.eln file', ['file'])
    if isinstance(reply, str):
      return False, 'Error publishing the file'
    return True, f'Published: {doi}, {res["persistentUrl"]}'


  def getDatasetInfo(self, ds_persistent_id: str, version: str = ':latest-published') -> dict[Any, Any] | Any:
    """
    Fetch JSON representation of a dataset
    Args:
      ds_persistent_id (str): The identifier of the dataset
      version (str): The version of the dataset
        Note dataset versions can be referred to as:
          :draft the draft version, if any
          :latest either a draft (if exists) or the latest published version
          :latest-published the latest published version
          x.y a specific version, where x is the major version number and y is the minor version number
          x same as x.0

    Returns:
      JSON representation of the dataset for successful request, otherwise the error message is returned
    """
    resp = requests.get(
      f"{self.server_url}/api/datasets/:persistentId/versions/{version}?persistentId={ds_persistent_id}",
      params={'Accept': 'application/json'}, timeout=10,
      headers={'Content-Type': 'application/json', 'X-Dataverse-key': self.api_token})
    if resp.status_code == 200:
      return resp.json().get('data')
    return f"Error fetching JSON representation of dataset: {ds_persistent_id} Info: {resp.json()}"


  def getDatasetVersions(self, ds_persistent_id: str) -> dict[Any, Any] | Any:
    """
    Fetch the version list for dataset
    Args:
      ds_persistent_id (str): The identifier of the dataset

    Returns:
      Version list for the dataset for successful request, otherwise the error message is returned
    """
    resp = requests.get(
      f"{self.server_url}/api/datasets/:persistentId/versions?persistentId={ds_persistent_id}",
      params={'Accept': 'application/json'}, timeout=10,
      headers={'Content-Type': 'application/json', 'X-Dataverse-key': self.api_token})
    if resp.status_code == 200:
      return resp.json().get('data')
    return f"Error fetching version list for dataset: {ds_persistent_id} Info: {resp.json()}"


  def getDatasetLocks(self, ds_persistent_id: str) -> dict[Any, Any] | Any:
    """
    Fetches locks for a dataset

    Explanation:
        This method fetches the locks for a dataset identified by the persistent ID
        It makes an asynchronous HTTP GET request to retrieve the locks information and handles exceptions

    Args:
        ds_persistent_id (str): The persistent ID of the dataset to fetch locks for

    Returns:
        dict[Any, Any] | Any: A dictionary containing the locks information if successful, or an error message
    """
    resp = requests.get(
      f"{self.server_url}/api/datasets/:persistentId/locks?persistentId={ds_persistent_id}",
      params={'Accept': 'application/json'}, timeout=10,
      headers={'Content-Type': 'application/json', 'X-Dataverse-key': self.api_token})
    if resp.status_code == 200:
      return {'locks': resp.json().get('data')}
    return f"Error fetching locks for dataset: {ds_persistent_id}  Info: {resp.json()}"


  def getDatasetFiles(self, ds_persistent_id: str, version: str = ':latest-published') -> dict[Any, Any] | Any:
    """
    Fetch the file list for dataset
    Args:
      ds_persistent_id (str): The identifier of the dataset
      version (str): The version of the dataset
        Note dataset versions can be referred to as:
          :draft the draft version, if any
          :latest either a draft (if exists) or the latest published version
          :latest-published the latest published version
          x.y a specific version, where x is the major version number and y is the minor version number
          x same as x.0

    Returns:
      File list for the dataset for successful request, otherwise the error message is returned
    """
    resp = requests.get(
      f"{self.server_url}/api/datasets/:persistentId/versions/{version}/files?persistentId={ds_persistent_id}",
      params={'Accept': 'application/json'}, timeout=10,
      headers={'Content-Type': 'application/json', 'X-Dataverse-key': self.api_token})
    if resp.status_code == 200:
      return resp.json().get('data')
    return f"Error fetching file list for dataset: {ds_persistent_id}  Info: {resp.json()}"


  def getDatasetMetadata(self, ds_persistent_id: str, version: str = ':latest-published') -> dict[Any, Any] | Any:
    """
    Fetch the metadata block for dataset
    Args:
      ds_persistent_id (str): The identifier of the dataset
      version (str): The version of the dataset
        Note dataset versions can be referred to as:
          :draft the draft version, if any
          :latest either a draft (if exists) or the latest published version
          :latest-published the latest published version
          x.y a specific version, where x is the major version number and y is the minor version number
          x same as x.0

    Returns:
      Metadata block for the dataset for successful request, otherwise the error message is returned
    """
    resp = requests.get(
      f"{self.server_url}/api/datasets/:persistentId/versions/{version}/metadata?persistentId={ds_persistent_id}",
      params={'Accept': 'application/json'}, timeout=10,
      headers={'Content-Type': 'application/json', 'X-Dataverse-key': self.api_token})
    if resp.status_code == 200:
      return resp.json().get('data')
    return f"Error fetching metadata block for dataset: {ds_persistent_id}  Info: {resp.json()}"


  def deleteEmptyDataverse(self) -> str | Any:
    """
    Deletes an empty dataverse

    Returns:
      Message for successful request, otherwise the error message is returned
    """
    resp = requests.delete(f"{self.server_url}/api/dataverses/{self.identifier}", headers=self.headers, timeout=10)
    if resp.status_code == 200:
      return resp.json().get('data').get('message')
    return f"Error deleting dataverse,  Info: {resp.json()}"


  def deletePublishedDataset(self, ds_persistent_id: str) -> str | Any:
    """
    Deletes a published dataset
    Note: The method fails presently due to a bug in dataverse API
          Every time the API returns 403 errors saying the user needs to be superuser
    Args:
      ds_persistent_id (str): The identifier of the dataset

    Returns:
      Message for successful request, otherwise the error message is returned
    """
    resp = requests.delete(
      f"{self.server_url}/api/datasets/:persistentId/destroy/",
      params={'persistentId': ds_persistent_id},
      headers=self.headers, timeout=10)
    if resp.status_code == 200:
      return resp.json().get('data').get('message')
    return f"Error deleting dataset, Id: {ds_persistent_id}, Info: {resp.json()}"


  def deleteNonEmptyDataverse(self) -> str | Any:
    """
    Deletes a non-empty dataverse

    Note: The method fails presently due to a bug in dataverse API while invoking the delete_published_dataset method

    Returns:
      Message for successful request, otherwise the error message is returned
    """
    contents = self.getDataverseContent()
    for content in contents:
      if content.get('type') == 'dataset':
        self.deletePublishedDataset(f"{content.get('protocol')}:{content.get('authority')}/{content.get('identifier')}")
      elif content.get('type') == 'dataverse':
        self.deleteNonEmptyDataverse()
      else:
        logging.error('Unknown content type: %s while deleting dataverse: %s on server: %s ', content.type,
                      self.identifier, self.server_url, exc_info=True)
    resp = requests.delete(f"{self.server_url}/api/dataverses/{self.identifier}", headers=self.headers, timeout=10)
    if resp.status_code == 200:
      return resp.json().get('data').get('message')
    return f"Error deleting dataverse, Id: {self.identifier}, Info: {resp.json()}"


  def prepareMetadata(self, metadata:dict[str,Any]) -> dict[str,Any]:
    """
    Prepares the metadata for uploading

    Args:
        metadata (dict): The metadata to be prepared

    Returns:
        dict: The prepared metadata
    """
    author = metadata['author']
    fields = [{'typeName': 'title', 'value': metadata['title'], 'typeClass': 'primitive'},
              {'typeName': 'author', 'value': [{'authorName': {'value': f"{author['last']}, {author['first']}"},
                'authorIdentifier': {'value': author['orcid']},
                'authorAffiliation': {'value': author['organizations'][0]['organization']}}], 'typeClass': 'compound'},
              {'typeName': 'datasetContact', 'value': [{'datasetContactEmail': {'value': author['email']},
                'datasetContactName': {'value': f"{author['last']}, {author['first']}"}}], 'typeClass': 'compound'},
              {'typeName': 'keywords', 'value': metadata['keywords'], 'typeClass': 'primitive'},
              {'typeName': 'publicationDate', 'value': datetime.now().strftime('%Y-%m-%d'), 'typeClass': 'primitive'},
              {'typeName': 'dsDescription', 'value': [{'dsDescriptionValue': {'value': metadata['description']}}],
                'typeClass': 'compound'},
              {'typeName': 'subject', 'value': [metadata['category']], 'typeClass': 'controlledVocabulary'}
            ] + metadata['additional']
    return {'metadata': {'datasetVersion': {'metadataBlocks': {'citation': {'fields': fields}}}}}



#!/usr/bin/python3
"""TEST using the FULL set of python-requirements: create the default example that all installations create and verify it thoroughly """
import asyncio
import logging
import unittest
import warnings
from pasta_eln.miscTools import getRORIDLabel, getORCIDName
from pasta_eln.AddOns.definition_autofill import getFirstWikidataEntry
from pasta_eln.UI.definitions.terminology_lookup_service import TerminologyLookupService

class TestStringMethods(unittest.TestCase):
  """
  derived class for this test
  """
  def test_main(self):
    """
    main function
    """
    # initialization: create database, destroy on filesystem and database and then create new one
    warnings.filterwarnings('ignore', message='numpy.ufunc size changed')
    warnings.filterwarnings('ignore', message='invalid escape sequence')
    warnings.filterwarnings('ignore', category=ResourceWarning, module='PIL')
    warnings.filterwarnings('ignore', category=ImportWarning)
    for package in ['urllib3', 'requests', 'asyncio', 'PIL', 'matplotlib.font_manager']:
      logging.getLogger(package).setLevel(logging.WARNING)
    logging.info('Start 99 test')

    answer = getRORIDLabel('02nv7yv05')
    assert answer=='Forschungszentrum Jülich', f'RORID lookup failed: {answer}'
    answer = getRORIDLabel('042nb2s44')
    assert answer=='Massachusetts Institute of Technology', f'RORID lookup failed: {answer}'

    answer = getORCIDName('0000-0003-0930-082X')
    assert answer==('Steffen','Brinckmann'), f'ORCID lookup failed: {answer}'
    answer = getORCIDName('0000-0001-8940-2361')
    assert answer==('Ruth','Schwaiger'), f'ORCID lookup failed: {answer}'

    # Add-on form-auto.py cannot be tested since it requires secret API key
    answer = getFirstWikidataEntry('water')
    assert answer=='http://www.wikidata.org/entity/Q283', f'Wikidata lookup failed: {answer}'
    answer = getFirstWikidataEntry('glucose')
    assert answer=='http://www.wikidata.org/entity/Q37525', f'Wikidata lookup failed: {answer}'


    # pasta_eln/UI/definitions/terminology_lookup_service.py:
    tsService = TerminologyLookupService()
    for searchTerm in ['force','strain']:  #always use two terms
      answers = []
      successServices = set()
      event_loop = asyncio.new_event_loop()
      asyncio.set_event_loop(event_loop)
      if lookup_results := event_loop.run_until_complete(tsService.do_lookup(searchTerm)):
        for service in lookup_results:
          for result in service['results']:
            serviceShort = {'ontology_lookup_service':'OLS  ',
                            'tib_terminology_service':'TIB  ',
                            'wikipedia':'wiki ',
                            'wikidata':'wikiD'
                            }[service['name']]
            answers.append(f"{result['iri'][:30]:<30} {serviceShort}  {result['information'][:100]}")
            successServices.add(serviceShort.strip())
      listStr = '\n - '.join(answers)
      print(f'Search term: {searchTerm} \n - {listStr}')
      print('Services used:', ', '.join(sorted(successServices)),'\n')
      assert successServices=={'wikiD', 'wiki', 'TIB', 'OLS'}


    # elabFTW API is tested in testComplicated since it requires a secret API key. Hence do not test
    # - pasta_eln/UI/config/projectGroup.py;
    # - pasta_eln/backendWorker/elabFTWapi.py

    # pasta_eln/backendWorker/inputOutput.py cannot be tested since it requires a specific location to test, which is known at this point in time

    # zenodo and dataverse API are tested in testComplicated since they require secret API keys



""" Edit properties of a docType """
import string
from typing import Callable, Optional
import pandas as pd
import qtawesome as qta
from PySide6.QtCore import Slot
from PySide6.QtGui import QRegularExpressionValidator
from PySide6.QtWidgets import QComboBox, QDialog, QDialogButtonBox, QLabel, QLineEdit, QVBoxLayout
from ...fixedStringsJson import allIcons
from ..guiCommunicate import Communicate
from ..guiStyle import TextButton, widgetAndLayout, widgetAndLayoutForm
from ..messageDialog import showMessage


class DocTypeEditor(QDialog):
  """ Edit properties of a docType """
  def __init__(self, comm:Communicate, docType:str, callback:Optional[Callable[[str,str],None]]=None):
    """
    Initialization

    Args:
      comm (Communicate): communication channel
      doc (dict):  document to change / create
      callback (function): callback to allow sending the new doc-type
    """
    super().__init__()
    self.comm = comm
    self.comm.backendThread.worker.beSendSQL.connect(self.onGetData)
    self.docType = docType
    self.shortcuts:list[list[str]] = []
    self.callback = callback
    mainL = QVBoxLayout(self)
    _, self.mainForm = widgetAndLayoutForm(mainL)
    self.setWindowTitle('Edit item type properties')
    mainL.addStretch(1)
    buttonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Save | QDialogButtonBox.StandardButton.Cancel)
    buttonBox.clicked.connect(self.closeDialog)
    mainL.addWidget(buttonBox)
    # initialize communication if needed
    self.comm.uiSendSQL.emit([{'type':'get_df',
                               'cmd':'SELECT docType, shortcut from docTypes'}])
    if self.docType in self.comm.docTypesTitles:
      self.comm.uiSendSQL.emit([{'type':'get_df',
                                 'cmd':f'SELECT * from docTypes WHERE docType=="{self.docType}"'}])
      self.initialData = None
    else:
      self.initialData = ['']*6
      self.paint()


  @Slot(str, pd.DataFrame)
  def onGetData(self, cmd:str, data:pd.DataFrame) -> None:
    """ Handle data received from backend worker
    Args:
      cmd (str): command that was sent to backend
      data (pd.DataFrame): DataFrame containing the data
    """
    if cmd == 'SELECT docType, shortcut from docTypes':
      self.shortcuts = [j for i,j in data.values.tolist() if len(j)==1 and i!=self.docType]
    elif cmd.startswith('SELECT * from docTypes WHERE'):
      self.initialData = data.values.tolist()[0]
      self.paint()


  def paint(self) -> None:
    """ Paint the dialog with the initial data """
    if self.initialData is None:
      return
    self.row1 = QLineEdit(self.initialData[0])
    self.row1.setToolTip('Name of item type: lower case letters only. Must be unique')
    if self.docType:
      self.row1.setDisabled(True)
    else:
      self.row1.setValidator(QRegularExpressionValidator(r'(^[a-wyz][\w\/]{3,}$)'))
    self.mainForm.addRow(QLabel('DocType '), self.row1)

    self.row2 = QLineEdit(self.initialData[2])
    self.row2.setToolTip('Label that the user reads: it is suggested to start with upper case and end with s as in "Samples"')
    self.mainForm.addRow(QLabel('Label '), self.row2)

    row3W, row3L = widgetAndLayout('H', None, 's')
    # Label('type:', 'h2', row3L)
    # self.comboType = QComboBox()
    # self.comboType.addItems(set(i.split('.')[0] for i in allIcons))
    # row3L.addWidget(self.comboType)
    # Label('icon:', 'h2', row3L)
    self.comboIcon = QComboBox()
    self.comboIcon.addItem('')
    for icon in allIcons:
      self.comboIcon.addItem(qta.icon(icon, scale_factor=1.5), icon)
    self.comboIcon.setCurrentText(self.initialData[3])
    row3L.addWidget(self.comboIcon)
    if self.docType.startswith('x'):
      self.comboIcon.setDisabled(True)
    self.mainForm.addRow(QLabel('Icon '), row3W)

    self.row4 = QLineEdit(self.initialData[4])
    self.row4.setToolTip('One letter shortcut used with Ctrl-. Must be unique.')
    if self.docType.startswith('x'):
      self.row4.setDisabled(True)
    else:
      dataSet = set(string.ascii_lowercase).difference(self.shortcuts)
      self.row4.setValidator(QRegularExpressionValidator(f'^[{"".join(dataSet)}]$'))
    self.mainForm.addRow(QLabel('Shortcut '), self.row4)


  def closeDialog(self, btn:TextButton) -> None:
    """
    cancel or save entered data

    Args:
      btn (QButton): save or cancel button
    """
    if btn.text().endswith('Cancel'):
      self.reject()
    else:
      label    = self.row2.text()
      if not label or label in [v['title'] for _,v in self.comm.docTypesTitles.items()]:
        showMessage(self, 'Error', 'The label has to be used and cannot be used already by another type.', 'Critical')
        return
      icon     = '' if self.docType.startswith('x') else self.comboIcon.currentText()
      shortcut = '' if self.docType.startswith('x') else self.row4.text()
      if self.docType:                                                               # update existing docType
        cmd = f"UPDATE docTypes SET title='{label}', shortcut='{shortcut}', icon='{icon}' WHERE docType = "\
              f"'{self.docType}'"
        self.comm.uiSendSQL.emit([{'type':'one',  'cmd':cmd}])
        docType = self.docType
      else:                                                  # create new docType, with default schema entries
        docType = self.row1.text()
        if not docType or docType in self.comm.docTypesTitles:
          showMessage(self, 'Error', 'DocType name is not valid or already exists', 'Critical')
          return
        self.comm.uiSendSQL.emit([{'type':'one', 'cmd':'INSERT INTO docTypes VALUES (?, ?, ?, ?, ?, ?)',
                                   'list':[docType, '', label, icon, shortcut, '']},
                                  {'type':'one', 'cmd':'INSERT INTO docTypeSchema VALUES (?, ?, ?, ?, ?, ?, ?)',
                                   'list':[docType, '', '0', 'name', '', '', '']},
                                  {'type':'one', 'cmd':'INSERT INTO docTypeSchema VALUES (?, ?, ?, ?, ?, ?, ?)',
                                   'list':[docType, '', '1', 'tags', '', '', '']},
                                  {'type':'one', 'cmd':'INSERT INTO docTypeSchema VALUES (?, ?, ?, ?, ?, ?, ?)',
                                   'list':[docType, '', '2', 'comment', '', '', '']}])
        if self.callback is not None:
          self.callback(label, docType)
      self.accept()
    return


  def reject(self) -> None:
    """ Reject the dialog, stop the thread and disconnect signals """
    self.comm.backendThread.worker.beSendSQL.disconnect(self.onGetData)
    super().reject()


  def accept(self) -> None:
    """ Accept the dialog, stop the thread and disconnect signals """
    self.comm.backendThread.worker.beSendSQL.disconnect(self.onGetData)
    super().accept()



""" widget that shows the details of the items """
import logging
import re
from enum import Enum
from pathlib import Path
from typing import Any
import pandas as pd
from PySide6.QtCore import Qt, Slot
from PySide6.QtWidgets import QLabel, QLayout, QScrollArea, QTextEdit
from ..backendWorker.worker import Task
from ..fixedStringsJson import SORTED_DB_KEYS, cssStyleHtmlEditors, defaultDataHierarchyNode
from ..textTools.handleDictionaries import dict2ul
from ..textTools.stringChanges import markdownEqualizer, tuple2html
from ._contextMenu import CommandMenu, executeContextMenu, initContextMenu
from .guiCommunicate import Communicate
from .guiStyle import IconButton, Image, Label, TextButton, widgetAndLayout
from .messageDialog import showMessage


class Details(QScrollArea):
  """ widget that shows the details of the items """
  def __init__(self, comm:Communicate):
    super().__init__()
    self.comm = comm
    self.comm.changeDetails.connect(self.change)
    self.comm.backendThread.worker.beSendDoc.connect(self.onGetData)
    self.comm.backendThread.worker.beSendTable.connect(self.onGetTable)
    self.comm.testExtractor.connect(self.testExtractor)
    self.doc:dict[str,Any]  = {}
    self.docID= ''
    self.idsTypesNames = pd.DataFrame(columns=['id','type','name'])
    self.textEditors:list[QTextEdit] = []

    # GUI elements
    self.mainW, self.mainL = widgetAndLayout('V', None)
    self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOn)
    self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
    self.setWidgetResizable(True)
    self.setWidget(self.mainW)

    headerW, self.headerL = widgetAndLayout('H', self.mainL, spacing='m', top='s', right='s')
    headerW.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
    headerW.customContextMenuRequested.connect(lambda pos: initContextMenu(self, pos))
    self.labelW = Label('','h1', self.headerL)
    self.headerL.addStretch(1)
    IconButton('mdi.file-tree-outline', self, [Command.TO_PROJECT], self.headerL, tooltip='Change to project')
    IconButton('fa5s.times-circle',     self, [Command.CLOSE],      self.headerL, tooltip='Close')
    self.specialW, self.specialL = widgetAndLayout('V', self.mainL, top='s')
    self.specialW.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
    self.specialW.customContextMenuRequested.connect(lambda pos: initContextMenu(self, pos))
    self.btnDetails = TextButton('Details', self, [Command.SHOW, 'Details'], self.mainL, \
                                 'Show / hide details', checkable=True, style='margin-top: 3px')
    self.metaDetailsW, self.metaDetailsL  = widgetAndLayout('V', self.mainL)
    self.btnVendor = TextButton('Vendor metadata', self, [Command.SHOW, 'Vendor'], self.mainL, \
                                'Show / hide vendor metadata', checkable=True, style='margin-top: 15px')
    self.metaVendorW, self.metaVendorL = widgetAndLayout('V', self.mainL)
    self.btnUser = TextButton('User metadata', self, [Command.SHOW, 'User'], self.mainL, \
                              'Show / hide user metadata', checkable=True, style='margin-top: 15px')
    self.metaUserW, self.metaUserL     = widgetAndLayout('V', self.mainL)
    self.btnDatabase = TextButton('ELN details', self, [Command.SHOW,'Database'], self.mainL, \
                                  'Show / hide database details', checkable= True, style='margin-top: 15px')
    self.metaDatabaseW, self.metaDatabaseL = widgetAndLayout('V', self.mainL)
    self.mainL.addStretch(1)


  @Slot(str)
  def change(self, docID:str) -> None:
    """ What happens when user clicks to change to a different document
    Args:
      docID (str): document-id
    """
    self.docID = docID
    self.comm.uiRequestDoc.emit(self.docID)


  @Slot(dict)
  def onGetData(self, doc:dict[str,Any]) -> None:
    """ Function to handle the received data
    Args:
      doc (dict): dictionary containing the document data
    """
    if 'id' in doc and doc['id'] == self.docID:
      self.doc = doc
      self.paint()


  @Slot(pd.DataFrame, str)
  def onGetTable(self, data:pd.DataFrame, docType:str) -> None:
    """
    Callback function to handle the received data

    Args:
      data (pd.DataFrame): DataFrame containing table
      docType (str): document type
    """
    data = data[['id','name']]
    data['type']= docType
    for _, row in data.iterrows():
      id_ = row['id']
      if id_ in self.idsTypesNames['id'].values:                            # Replace the row with matching id
        self.idsTypesNames.loc[self.idsTypesNames['id']==id_, ['name','type']] = row[['name', 'type']].values
      else:                                                                              # Concatenate new row
        self.idsTypesNames = pd.concat([self.idsTypesNames, pd.DataFrame([row])], ignore_index=True)
    self.paint()


  @Slot()
  def paint(self) -> None:
    """ What happens when details should change """
    if self.isHidden():
      return
    # Delete old widgets from layout
    for i in reversed(range(self.metaDetailsL.count())):
      self.metaDetailsL.itemAt(i).widget().setParent(None)
    for i in reversed(range(self.metaVendorL.count())):
      self.metaVendorL.itemAt(i).widget().setParent(None)
    for i in reversed(range(self.metaUserL.count())):
      self.metaUserL.itemAt(i).widget().setParent(None)
    for i in reversed(range(self.metaDatabaseL.count())):
      self.metaDatabaseL.itemAt(i).widget().setParent(None)
    for i in reversed(range(self.specialL.count())):
      self.specialL.itemAt(i).widget().setParent(None)
    self.specialW.hide()
    self.metaDetailsW.hide()
    self.metaVendorW.hide()
    self.metaUserW.hide()
    self.metaDatabaseW.hide()
    self.btnDetails.setChecked(True)
    self.btnVendor.setChecked(True)
    self.btnUser.setChecked(True)
    self.btnDatabase.setChecked(False)
    self.textEditors = []
    # Create new
    if not self.docID or not self.doc:
      self.hide()
      return
    if self.doc['type'][0] not in self.comm.docTypesTitles:
      dataHierarchyNode = defaultDataHierarchyNode
    else:
      dataHierarchyNode = self.comm.dataHierarchyNodes[self.doc['type'][0]]
    self.labelW.setText(self.doc['name'] if len(self.doc['name'])<80 else self.doc['name'][:77]+'...')
    if 'metaVendor' not in self.doc:
      self.btnVendor.hide()
    if 'metaUser' not in self.doc:
      self.btnUser.hide()
    for key in self.doc:
      if key=='image':
        Image(self.doc['image'], self.specialL, anyDimension=self.width()-20)
        self.specialW.show()
      elif key=='content':
        text = QTextEdit()                                                   # pylint: disable=qt-local-widget
        text.setMarkdown(self.doc['content'])
        size = self.comm.configuration['GUI']['imageSizeDetails'] if hasattr(self.comm, 'configuration') else 300
        text.setFixedHeight(int(size/3*2))
        text.setReadOnly(True)
        self.textEditors.append(text)
        self.specialL.addWidget(text)
        self.specialW.show()
      elif key in ['name']:                                                                              #skip
        continue
      elif key in SORTED_DB_KEYS:
        self.addDocDetails(self.metaDatabaseL, key, self.doc[key], dataHierarchyNode)
        self.btnDatabase.setChecked(False)
      elif key=='metaVendor':
        self.btnVendor.show()
        self.addDocDetails(self.metaVendorL,   '',  self.doc[key], dataHierarchyNode)
        self.metaVendorW.show()
      elif key=='metaUser':
        self.btnUser.show()
        self.addDocDetails(self.metaUserL,     '',  self.doc[key], dataHierarchyNode)
        self.metaUserW.show()
      else:
        self.addDocDetails(self.metaDetailsL,  key, self.doc[key], dataHierarchyNode)
        self.metaDetailsW.show()
    return


  def execute(self, command:list[Any]) -> None:
    """
    Hide / show the widget underneath the button

    Args:
      command (list): area to show/hide
    """
    if command[0] is Command.SHOW:
      if getattr(self, f'btn{command[1]}').isChecked():                                #get button in question
        getattr(self, f'meta{command[1]}W').show()
      else:
        getattr(self, f'meta{command[1]}W').hide()
    elif command[0] is Command.CLOSE:
      self.hide()
    elif command[0] is Command.TO_PROJECT:
      self.comm.changeProject.emit(self.doc['branch'][0]['stack'][0], self.doc['id'])
    elif isinstance(command[0], CommandMenu):
      if executeContextMenu(self, command):
        self.comm.changeTable.emit('','')
        self.comm.changeDetails.emit(self.doc['id'])
    else:
      logging.error('Details command unknown: %s',command, exc_info=True)
    return


  @Slot()
  def testExtractor(self) -> None:
    """
    User selects to test extractor on this dataset
    """
    logging.debug('details:testExtractor')
    if len(self.doc)>1:
      pathStr = self.doc['branch'][0]['path']
      if pathStr is None:
        showMessage(self, 'Warning', 'Selected item has no path.')
      else:
        path = Path(pathStr)
        if not path.as_posix().startswith('http'):
          path = self.comm.basePath/path
        self.comm.uiRequestTask.emit(Task.EXTRACTOR_TEST, {'fileName':str(path), 'style':'html', 'recipe':'', 'saveFig':''})
    else:
      showMessage(self, 'Warning', 'No item was selected via list-view, i.e. no details are shown.')
    return


  def resizeWidth(self, width:int) -> None:
    """ called if details is resized by splitter at the parent widget
    - resize all text-documents
    """
    if not self.textEditors:
      return
    self.metaDetailsW.setFixedWidth(width)
    for text in self.textEditors:
      text.document().setTextWidth(width-80)
      height:int = text.document().size().toTuple()[1]                                    # type:ignore[index]
      text.setFixedHeight(height)
    return



  def addDocDetails(self, layout:QLayout, key:str, value:Any, dataHierarchyNode:list[dict[str,Any]]) -> str:
    """ add document details to a widget's layout: take care of all the formatting

    Args:
      layout (QLayout): layout to which to add
      key    (str): key/label to add
      value  (Any): information
      dataHierarchyNode (list): information on data-structure

    Returns:
      str: /n separated lines of text
    """
    if not key and isinstance(value,dict):
      return '\n'.join([self.addDocDetails(layout, k, v, dataHierarchyNode) for k, v in value.items()])
    if not value:
      return ''
    link = False
    labelStr = ''
    if key=='tags':
      rating = ['\u2605'*int(i[1]) for i in value if re.match(r'^_\d$', i)]
      tags = [i for i in value if not re.match(r'^_\d$', i)]
      labelStr = f'Rating: {rating[0]}' if rating else ''
      labelStr = f'{labelStr}   Tags: '+' '.join(tags)
      if layout is not None:
        label = QLabel(labelStr)
        label.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)
        layout.addWidget(label)
    elif (isinstance(value,str) and '\n' in value) or key=='comment':        # long values with /s or comments
      labelW, labelL = widgetAndLayout('H', layout, top='s', bottom='s')
      labelL.addWidget(QLabel(f'{key}: '), alignment=Qt.AlignmentFlag.AlignTop)
      text = QTextEdit()                                                     # pylint: disable=qt-local-widget
      text.setMarkdown(markdownEqualizer(value))
      bgColor = self.comm.palette.get('secondaryDark', 'background-color')
      fgColor = self.comm.palette.get('secondaryText', 'color')
      text.setStyleSheet(f"QTextEdit {{ border: none; padding: 0px; {bgColor} {fgColor}}}")
      text.document().setTextWidth(labelW.width())
      if hasattr(self, 'rescaleTexts'):
        self.textEditors.append(text)
      height:int = text.document().size().toTuple()[1]                                    # type:ignore[index]
      text.setFixedHeight(height)
      text.setReadOnly(True)
      self.textEditors.append(text)
      labelL.addWidget(text, stretch=1)
    else:
      dataHierarchyItems = [dict(i) for i in dataHierarchyNode if i['name']==key]
      docID = ''
      if len(dataHierarchyItems)==1 and 'list' in dataHierarchyItems[0] and dataHierarchyItems[0]['list'] and \
          ',' not in dataHierarchyItems[0]['list'] and ' ' not  in dataHierarchyItems[0]['list']:#choice among docType
        if dataHierarchyItems[0]['list'] not in self.idsTypesNames['type'].values:
          self.comm.uiRequestTable.emit(dataHierarchyItems[0]['list'], '', True)
        else:
          names= list(self.idsTypesNames[self.idsTypesNames.id==value[0]]['name'])
          if len(names)==1:                                            # default find one item that we link to
            docID = value[0]
            value = '\u260D '+names[0]
            link = True
          elif not names:
            value = value[0] if isinstance(value,tuple) else value
          else:
            raise ValueError(f'list target exists multiple times. Key: {key}')
      elif isinstance(value, list):
        value = ', '.join([str(i) for i in value])
      labelStr = f'<b>{key.capitalize()}</b>: {value}'
      if isinstance(value, tuple) and len(value)==4:
        k,v = tuple2html(key, value)
        labelStr = f'{k.capitalize()}: {v}<br>'
      if isinstance(value, dict):
        newValue = {}
        for k,v in value.items():
          if isinstance(v, tuple) and len(v)==4:
            k2,v2 = tuple2html(k, v)
            newValue[k2] = v2
          elif isinstance(v, (list, tuple)):
            newValue[k] = v[0]
          else:
            newValue[k] = v
        labelStr = f'{cssStyleHtmlEditors}{key.capitalize()}: {dict2ul(newValue)}'
      if layout is not None:
        label = Label(labelStr, function=lambda x,y: self.clickLink(x,y) if link else None, docID=docID)
        label.setOpenExternalLinks(True)
        label.setWordWrap(True)
        layout.addWidget(label)
    return labelStr


  def clickLink(self, label:str, docID:str) -> None:
    """
    Click link in details

    Args:
      label (str): label on link
      docID (str): docID to which to link
    """
    logging.debug('used link on %s|%s',label,docID)
    self.comm.changeDetails.emit(docID)
    return


class Command(Enum):
  """ Commands used in this file """
  SHOW             = 1
  CLOSE            = 2
  TO_PROJECT       = 3



""" Main class of config tab on parameters (e.g. API keys) for add-ons """
import importlib
import json
from enum import Enum
from pathlib import Path
from typing import Callable
from PySide6.QtWidgets import QApplication, QDialog, QGroupBox, QLineEdit, QVBoxLayout
from ...fixedStringsJson import CONF_FILE_NAME
from ..guiCommunicate import Communicate
from ..guiStyle import Label, TextButton, widgetAndLayout
from ..messageDialog import showMessage


class ConfigurationAddOnParameter(QDialog):
  """ Main class of config tab on authors
  """
  def __init__(self, comm:Communicate, callbackFinished:Callable[[bool],None]):
    """
    Initialization

    Args:
      comm (Communicate): communication
      callbackFinished (function): callback function to call upon end
    """
    super().__init__()
    self.comm = comm
    self.callbackFinished = callbackFinished
    mainL = QVBoxLayout(self)
    Label('Define Add-On parameters','h2',mainL)

    #GUI elements
    self.allLineEdits:list[tuple[str,str,str,QLineEdit]] = []
    self.allGroupBoxes = []
    if hasattr(comm, 'configuration'):
      addOns = comm.configuration['projectGroups'][comm.projectGroup]['addOns']
      for addOnType in addOns:                                                        # loop over add-on types
        if addOnType != 'extractors' and addOns[addOnType]:
          for name, _ in addOns[addOnType].items():                                        # loop over add-ons
            groupbox = QGroupBox(name.capitalize())
            mainL.addWidget(groupbox)
            groupLayout = QVBoxLayout(groupbox)
            self.allGroupBoxes.append((addOnType, name, groupbox, groupLayout))

    #final button box
    mainL.addStretch(1)
    _, buttonLineL = widgetAndLayout('H', mainL, 'm')
    tooltip = 'Scan files to find parameters. Takes time.'
    self.scanBtn = TextButton('Scan',                self, [Command.SCAN],   buttonLineL, tooltip)
    self.scanBtn.setStyleSheet('background: orange; color: black;')
    buttonLineL.addStretch(1)
    self.saveBtn = TextButton('Save', self, [Command.SAVE],   buttonLineL, 'Save changes')
    self.saveBtn.setShortcut('Ctrl+Return')
    self.cancelBtn = TextButton('Cancel',              self, [Command.CANCEL], buttonLineL, 'Discard changes')


  def execute(self, command:list[str]) -> None:
    """
    Execute a command

    Args:
      command (list[str]): command to execute
    """
    if command[0] is Command.CANCEL:
      self.reject()
      self.callbackFinished(False)
    elif command[0] is Command.SAVE:
      apiKeys = self.comm.configuration.get('addOnParameter',{})
      for _, name, param, lineEdit in self.allLineEdits:
        if name not in apiKeys:
          apiKeys[name] = {}
        apiKeys[name][param] = lineEdit.text()
      self.comm.configuration['addOnParameter'] = apiKeys
      with open(Path.home()/CONF_FILE_NAME, 'w', encoding='utf-8') as fConf:
        fConf.write(json.dumps(self.comm.configuration,indent=2))
      self.accept()
      self.callbackFinished(False)
    elif command[0] is Command.SCAN:
      self.allLineEdits = []
      for addonType, name, groupbox, groupLayout in self.allGroupBoxes:
        QApplication.processEvents()                                                        # Force GUI update
        try:
          module      = importlib.import_module(name)         # ISSUE: slow since imports all dependencies,...
          requiredParam = module.reqParameter
          try:
            helpText = module.helpText
          except AttributeError:
            helpText = ''
          if not requiredParam:
            groupbox.hide()
          for param, tooltip in requiredParam.items():                                  # loop over parameters
            _, barL = widgetAndLayout('H', groupLayout, 'm', 's', 's', 's', 's')
            Label(f'{name}.py: {param}', 'h4', barL, tooltip=tooltip)
            lineEdit = QLineEdit()                                           # pylint: disable=qt-local-widget
            barL.addWidget(lineEdit)
            if helpText:
              TextButton('?', self, command=[helpText], layout=barL)
            self.allLineEdits.append((addonType,name,param,lineEdit))
        except Exception:
          Label(f'{name}.py: Error occurred; please check add-on.', 'h4', groupLayout)
    else:
      showMessage(self, 'Help', command[0], 'Information')
    return


class Command(Enum):
  """ Commands used in this file """
  SCAN        = 1
  SAVE        = 2
  CANCEL      = 3



""" Widget that shows the content of project in a electronic labnotebook """
import logging
import os
from enum import Enum
from typing import Any, Optional
from anytree import Node, PreOrderIter
from PySide6.QtCore import QItemSelectionModel, QModelIndex, Qt, Slot
from PySide6.QtGui import QAction, QStandardItem, QStandardItemModel
from PySide6.QtWidgets import QLabel, QMenu, QMessageBox, QTextEdit, QVBoxLayout, QWidget
from ..backendWorker.worker import Task
from ..fixedStringsJson import DO_NOT_RENDER
from ..miscTools import callAddOn
from ..textTools.handleDictionaries import doc2markdown
from ..textTools.stringChanges import createDirName
from .guiCommunicate import Communicate
from .guiStyle import Action, Label, TextButton, widgetAndLayout
from .messageDialog import showMessage
from .projectTreeView import TreeView


class Project(QWidget):
  """ Widget that shows the content of project in a electronic labnotebook """
  def __init__(self, comm:Communicate):
    super().__init__()
    self.comm = comm
    self.comm.changeProject.connect(self.change)
    self.comm.backendThread.worker.beSendHierarchy.connect(self.onGetData)
    self.hierarchy = Node('__none__')
    self.docProj:dict[str,Any] = {}
    self.projID = ''
    self.docIDHighlight = ''
    self.showAll= self.comm.configuration['GUI']['showHidden']=='Yes'

    self.mainL = QVBoxLayout()
    self.setLayout(self.mainL)
    self.tree :Optional[TreeView]            = None
    self.model:Optional[QStandardItemModel]  = None
    self.allDetails                          = QTextEdit()
    self.actHideDetail                       = QAction()
    self.actionFoldAll                       = QAction()
    self.showDetailsAll = False
    self.btnAddSubfolder:Optional[TextButton] = None
    self.btnEditProject:Optional[TextButton]  = None
    self.btnMore:Optional[TextButton]         = None
    self.btnVisibility:Optional[TextButton]   = None
    self.lineSep = 20
    self.META_ROLE = Qt.ItemDataRole.UserRole + 1


  @Slot(Node, dict)
  def onGetData(self, hierarchy:Node, doc:dict[str,Any]) -> None:
    """
    Callback function to handle the received data

    Args:
      hierarchy (Node): hierarchy of the project
      doc (pd.DataFrame): DataFrame containing table
    """
    if doc:
      self.docProj = doc
    self.hierarchy = hierarchy
    self.paint()


  @Slot(str, str)
  def change(self, projID:str, docID:str) -> None:
    """ Change project to projID and docID
    Args:
      projID (str): project ID
      docID (str): document ID
    """
    self.docIDHighlight = docID
    self.projID = projID
    self.comm.uiRequestHierarchy.emit(projID, self.showAll)


  def projHeader(self) -> None:
    """
    Initialize / Create header of page
    """
    # remove if still there
    for i in reversed(range(self.mainL.count())):                                                  #remove old
      self.mainL.itemAt(i).widget().setParent(None)
    logging.debug('ProjectView elements at 2: %i',self.mainL.count())
    if not self.docProj:
      return
    # TOP LINE includes name on left, buttons on right
    _, topLineL       = widgetAndLayout('H',self.mainL,'m')
    hidden, menuTextHidden = ('     \U0001F441', 'Mark project as shown') \
                       if [b for b in self.docProj['branch'] if False in b['show']] else \
                       ('', 'Mark project as hidden')
    topLineL.addWidget(Label(self.docProj['name']+hidden, 'h2'))
    showStatus = '(Show all items)' if self.showAll else '(Hide hidden items)'
    topLineL.addWidget(QLabel(showStatus))
    topLineL.addStretch(1)
    # buttons in top line
    self.btnAddSubfolder = TextButton('Add subfolder', self, [Command.ADD_CHILD], topLineL)
    self.btnEditProject =  TextButton('Edit project',  self, [Command.EDIT],      topLineL)
    self.btnVisibility = TextButton(  'Visibility',    self, [],                  topLineL)
    visibilityMenu = QMenu(self)
    self.actHideDetail = Action('Hide project details',self, [Command.SHOW_PROJ_DETAILS],visibilityMenu)
    menuTextItems = 'Hide hidden items' if self.showAll else 'Show hidden items'
    minimizeItems = 'Show all item details' if self.showDetailsAll else 'Hide all item details'
    Action( menuTextItems,    self, [Command.HIDE_SHOW_ITEMS],  visibilityMenu)
    Action( menuTextHidden,   self, [Command.HIDE],             visibilityMenu)
    self.actionFoldAll     = Action( minimizeItems,    self, [Command.SHOW_DETAILS],     visibilityMenu)
    self.btnVisibility.setMenu(visibilityMenu)
    self.btnMore = TextButton('More',           self, [], topLineL)
    moreMenu = QMenu(self)
    Action('Scan',                      self, [Command.SCAN], moreMenu)
    for docType, value in self.comm.docTypesTitles.items():
      if docType[0]!='x':
        icon = 'fa5s.asterisk' if value['icon']=='' else value['icon']
        Action(f'list of {value["title"].lower()}',   self, [Command.SHOW_TABLE, docType], moreMenu, icon=icon)
    Action('list of unidentified',     self, [Command.SHOW_TABLE, '-'],     moreMenu, icon='fa5.file')
    moreMenu.addSeparator()
    projectGroup = self.comm.configuration['projectGroups'][self.comm.projectGroup]
    if projectAddOns := projectGroup.get('addOns',{}).get('project',''):
      for label, description in projectAddOns.items():
        Action(description, self, [Command.ADD_ON, label], moreMenu)
      moreMenu.addSeparator()
    Action('Delete project',            self, [Command.DELETE], moreMenu)
    self.btnMore.setMenu(moreMenu)

    # Details section
    # self.infoW = QScrollArea()
    # self.infoW.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
    # self.infoW.setWidgetResizable(True)
    self.allDetails.setMarkdown(doc2markdown(self.docProj, DO_NOT_RENDER, self.comm.dataHierarchyNodes['x0'],
                                             self))
    if not self.docProj['gui'][0]:
      self.allDetails.hide()
      self.actHideDetail.setText('Show project details')
    self.allDetails.resizeEvent = self.commentResize                                            # type: ignore
    bgColor = self.comm.palette.get('secondaryDark', 'background-color')
    fgColor = self.comm.palette.get('secondaryText', 'color')
    #TODO: Debug/solve on windows: For none: no color is set, as it should; but then the Windows10 color is white not the default background
    self.allDetails.setStyleSheet(f"border: none; padding: 0px; {bgColor} {fgColor}")
    self.allDetails.setReadOnly(True)
    self.mainL.addWidget(self.allDetails)
    self.commentResize(None)
    return


  def commentResize(self, _:Any) -> None:
    """ called if comment is resized because widget initially/finally knows its size
    - comment widget is hard coded size it depends on the rendered size
    """
    if self.allDetails is None:
      return
    self.allDetails.document().setTextWidth(self.width()-20)
    height:int = self.allDetails.document().size().toTuple()[1]                          # type: ignore[index]
    self.allDetails.setMaximumHeight(height+12)
    self.allDetails.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
    return


  def paint(self) -> None:
    """
    What happens when user clicks to change project that is shown
    """
    if self.isHidden() and 'PYTEST_CURRENT_TEST' not in os.environ:
      return
    #initialize
    for i in reversed(range(self.mainL.count())):                                                  #remove old
      self.mainL.itemAt(i).widget().setParent(None)
    logging.debug('ProjectView elements at 1: %i',self.mainL.count())
    selectedIndex = None
    self.model = QStandardItemModel()
    self.tree = TreeView(self, self.comm, self.model)
    # self.tree.setSelectionBehavior(QAbstractItemView.SelectRows)
    # self.tree.setSelectionMode(QAbstractItemView.SingleSelection)
    self.model.itemChanged.connect(self.modelChanged)
    rootItem = self.model.invisibleRootItem()
    #Populate model body of change project: start recursion
    if self.hierarchy is not None and self.hierarchy.name == '__ERROR_in_getHierarchy__':
      showMessage(self, 'Error', 'There is an error in the project hierarchy: a parent of a node is incorrect.', 'Critical')
      return
    try:
      for node in PreOrderIter(self.hierarchy, maxlevel=2):
        if node is None or node.is_root:                                                      # Project header
          self.projHeader()
        else:
          rootItem.appendRow(self.iterateTree(node))
    except AttributeError:
      self.tree = TreeView(self, self.comm, QStandardItemModel())    # if hierarchy is None, create empty tree
    # collapse / expand depending on stored value
    # by iterating each leaf, and converting item and index
    root = self.model.invisibleRootItem()
    self.setExpandedState(root)
    if selectedIndex is not None:
      self.tree.selectionModel().select(selectedIndex, QItemSelectionModel.Select)
      self.tree.setCurrentIndex(selectedIndex)
    self.mainL.addWidget(self.tree)
    logging.debug('ProjectView elements at 4: %i',self.mainL.count())
    if self.hierarchy is not None and len(self.hierarchy.children)>0 and self.btnAddSubfolder is not None:
      self.btnAddSubfolder.setVisible(False)
    self.tree.expanded.connect(lambda index: self.actionExpandCollapse(index, True))
    self.tree.collapsed.connect(lambda index: self.actionExpandCollapse(index, False))
    if self.docIDHighlight:
      self.tree.scrollToDoc(self.docIDHighlight)
      self.docIDHighlight = ''                                                         # reset after scrolling
    return


  def setExpandedState(self, node:QStandardItem) -> None:
    """ Recursive function to set the expanded state of nodes

    Args:
      node (QStandardItem): node to process
    """
    if self.model is None or self.tree is None:
      return
    for iRow in range(node.rowCount()):
      item = node.child(iRow)
      data = item.data(self.META_ROLE)
      if data['hierStack'].split('/')[-1][0]=='x':
        index = self.model.indexFromItem(item)
        self.tree.setExpanded(index, data['gui'][1])
      self.setExpandedState(item)
    return


  def actionExpandCollapse(self, index:QModelIndex, flag:bool) -> None:
    """ Action upon expansion or collapsing of folder (showing its children)

    Args:
      index (QModelIndex): index that send the signal
      flag (bool): true=expand=show-children, false=collapse=hide-children
    """
    if self.model is None:
      return
    meta = index.data(self.META_ROLE)
    if not isinstance(meta, dict):
      return
    gui  = [meta['gui'][0]]+[flag]
    docID = meta['hierStack'].split('/')[-1]
    self.model.itemFromIndex(index).setData({ **meta, **{'gui':gui}}, self.META_ROLE)
    self.comm.uiRequestTask.emit(Task.SET_GUI, {'docID':docID, 'gui':gui})
    return


  def execute(self, command:list[Any]) -> None:
    """
    Event if user clicks button in the center

    Args:
      command (list): list of commands
    """
    if command[0] is Command.EDIT:
      self.comm.formDoc.emit({'id':self.docProj['id']})
      self.change(self.projID,'')
      #collect information and then change
      oldPath = self.comm.basePath/self.docProj['branch'][0]['path']
      if oldPath.is_dir():
        newPath = self.comm.basePath/createDirName(self.docProj, 0, self.comm.basePath)
        if oldPath != newPath:
          oldPath.rename(newPath)
      self.comm.changeSidebar.emit('redraw')
    elif command[0] is Command.DELETE:
      ret = QMessageBox.critical(self, 'Warning', 'Are you sure you want to delete project?', \
                                 QMessageBox.StandardButton.No | QMessageBox.StandardButton.Yes,
                                 QMessageBox.StandardButton.No)
      if ret==QMessageBox.StandardButton.Yes:
        self.comm.uiRequestTask.emit(Task.DELETE_DOC, {'docID':self.projID})
        #update sidebar, show projects
        self.comm.changeTable.emit('x0','')
    elif command[0] is Command.SCAN:
      self.comm.uiRequestTask.emit(Task.SCAN, {'docID':self.projID})
      self.comm.changeProject.emit(self.projID,'')
    elif command[0] is Command.SHOW_PROJ_DETAILS:
      self.docProj['gui'][0] = not self.docProj['gui'][0]
      self.comm.uiRequestTask.emit(Task.SET_GUI, {'docID':self.projID, 'gui':self.docProj['gui']})
      if self.allDetails is not None and self.allDetails.isHidden():
        self.allDetails.show()
        self.actHideDetail.setText('Hide project details')
      elif self.allDetails is not None:
        self.allDetails.hide()
        self.actHideDetail.setText('Show project details')
    elif command[0] is Command.HIDE:
      self.comm.uiRequestTask.emit(Task.HIDE_SHOW, {'docID':self.projID})
      self.comm.uiRequestHierarchy.emit(self.projID, self.showAll)
      self.comm.changeSidebar.emit('')
    elif command[0] is Command.SHOW_DETAILS and self.tree is not None:
      def recursiveRowIteration(index:QModelIndex) -> None:
        for subRow in range(self.tree.model().rowCount(index)):                     # type: ignore[union-attr]
          subIndex = self.tree.model().index(subRow,0, index)                       # type: ignore[union-attr]
          subItem  = self.tree.model().itemFromIndex(subIndex)                      # type: ignore[union-attr]
          meta = subItem.data(self.META_ROLE)
          if not isinstance(meta, dict):
            continue
          docID    = meta['hierStack'].split('/')[-1]
          gui      = meta['gui']
          gui[0]   = self.showDetailsAll
          subItem.setData({ **meta, **{'gui':gui}}, self.META_ROLE)
          self.comm.uiRequestTask.emit(Task.SET_GUI, {'docID':docID, 'gui':gui})
          recursiveRowIteration(subIndex)
      recursiveRowIteration(self.tree.model().index(-1,0))
      self.showDetailsAll = not self.showDetailsAll
      if self.showDetailsAll:
        self.actionFoldAll.setText('Show all item details')
      else:
        self.actionFoldAll.setText('Hide all item details')
    elif command[0] is Command.HIDE_SHOW_ITEMS:
      self.showAll = not self.showAll
      self.change('','')
    elif command[0] is Command.ADD_CHILD:
      self.comm.uiRequestTask.emit(Task.ADD_DOC, {'hierStack':[self.projID], 'docType':'x1', 'doc':{'name':'new item'}})
      self.comm.uiRequestHierarchy.emit(self.projID, self.showAll)

    elif command[0] is Command.SHOW_TABLE:
      self.comm.changeTable.emit(command[1], self.projID)
    elif command[0] is Command.ADD_ON:
      callAddOn(command[1], self.comm, self.projID, self)
    else:
      logging.error('Project menu unknown: %s',command, exc_info=True)
    return


  def modelChanged(self, item:QStandardItem) -> None:
    """
    Autocalled after drag-drop and other changes, record changes to backend and database directly

    Args:
      item (QStandardItem): item changed, new location
    """
    meta = item.data(self.META_ROLE)
    if not isinstance(meta, dict):
      return
    # gather old information
    stackOld = meta['hierStack'].split('/')[:-1]
    docID    = meta['hierStack'].split('/')[-1]
    childOld = meta['childNum']
    # gather new information
    stackNew = []                                                                             #create reversed
    currentItem = item
    while currentItem.parent() is not None:
      currentItem = currentItem.parent()
      metaParent = currentItem.data(self.META_ROLE)
      docIDj = metaParent['hierStack'].split('/')[-1]
      stackNew.append(docIDj)
    stackNew = [self.projID] + stackNew[::-1]                                      #add project id and reverse
    childNew = item.row()
    # compare
    logging.debug('Change project: docID %s | old stack %s child %i | new stack %s child %i'\
                  , docID, str(stackOld), childOld, str(stackNew), childNew)
    if stackOld==stackNew and childOld==childNew:                                #nothing changed, just redraw
      return
    self.comm.uiRequestTask.emit(Task.MOVE_LEAVES, {'docID':docID, 'stackOld':stackOld, 'stackNew':stackNew,
                                                    'childOld':childOld, 'childNew':childNew})
    item.setData(item.data() | {'hierStack': '/'.join(stackNew+[docID]), 'childNum':childNew})
    return

  def iterateTree(self, nodeHier:Node) -> QStandardItem:
    """
    Recursive function to translate the hierarchical node into a tree-node

    Args:
      nodeHier (Anytree.Node): anytree node

    Returns:
      QtTreeWidgetItem: tree node
    """
    #prefill docID
    hierStack = '/'.join([i.id for i in nodeHier.ancestors]+[nodeHier.id])
    gui = nodeHier.gui
    nodeTree = QStandardItem(nodeHier.name)
    nodeTree.setData({'hierStack':hierStack, 'docType':nodeHier.docType, 'gui':gui, 'childNum':nodeHier.childNum}, self.META_ROLE)
    if nodeHier.id[0]=='x':
      nodeTree.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsDragEnabled | Qt.ItemIsDropEnabled)# type: ignore
    else:
      nodeTree.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsDragEnabled)          # type: ignore
    children = []
    for childHier in nodeHier.children:
      childTree = self.iterateTree(childHier)
      children.append(childTree)
    if children:
      nodeTree.appendRows(children)
    return nodeTree


class Command(Enum):
  """ Commands used in this file """
  EDIT   = 1
  DELETE = 2
  SCAN   = 3
  HIDE   = 4
  SHOW_PROJ_DETAILS = 5
  HIDE_SHOW_ITEMS   = 6
  SHOW_DETAILS      = 7
  ADD_CHILD         = 8
  SHOW_TABLE        = 9
  ADD_ON            = 10



""" Upload for Zenodo and Dataverse """
import json
from enum import Enum
from pathlib import Path
from typing import Any
import qtawesome as qta
from PySide6.QtCore import Qt, Slot
from PySide6.QtWidgets import QCheckBox, QDialog, QLabel, QLineEdit, QVBoxLayout
from ...backendWorker.worker import Task
from ...fixedStringsJson import CONF_FILE_NAME
from ..guiCommunicate import Communicate
from ..guiStyle import Label, TextButton, widgetAndLayout, widgetAndLayoutGrid
from ..messageDialog import showMessage


class UploadGUI(QDialog):
  """ Upload for Zenodo and Dataverse """
  def __init__(self, comm:Communicate):
    """
    Initialization

    Args:
      backend (Backend): PASTA-ELN backend
      callbackFinished (function): callback function to call upon end
    """
    super().__init__()
    self.comm    = comm
    self.comm.backendThread.worker.beSendDoc.connect(self.onGetData)
    self.docProject:dict[str,Any] = {}
    self.setWindowTitle('Upload project to repository')

    # GUI elements
    self.mainL = QVBoxLayout(self)
    self.leTitle                       = QLineEdit()
    self.leDescription                 = QLineEdit()
    self.leKeywords                    = QLineEdit()
    self.leCategory                    = QLineEdit()
    self.leAdditional                  = QLineEdit()
    self.allDocTypes:list[list[str]]   = []
    self.allCheckboxes:list[QCheckBox] = []

    if not self.comm.projectID:
      showMessage(self, 'Error', 'You have to have an open project to upload', 'Critical')
      return
    self.comm.uiRequestDoc.emit(self.comm.projectID)


  @Slot(dict)
  def onGetData(self, doc:dict[str,Any]) -> None:
    """
    Callback function to handle the received data

    Args:
      doc (dict): project document
    """
    if self.comm.projectID and doc['id']==self.comm.projectID:
      self.docProject = doc
      self.paint()


  def paint(self) -> None:
    """ Paint the dialog with the current data """
    if not self.docProject:
      return
    Label('Upload to a repository', 'h1', self.mainL)
    _, center = widgetAndLayout('H', self.mainL, spacing='l', bottom='l', top='m')
    repositories = self.comm.configuration['repositories']
    leftSideW, leftSide = widgetAndLayoutGrid(center, spacing='m', right='l')
    leftSideW.setStyleSheet('border-right: 2px solid black;')
    leftSide.setAlignment(Qt.AlignTop)                                            # type: ignore[attr-defined]
    leftSide.addWidget(Label('Metadata','h2'), 0, 0)
    leftSide.addWidget(QLabel('Title'), 1, 0)
    self.leTitle = QLineEdit(self.docProject['name'])
    self.leTitle.setMinimumWidth(400)
    leftSide.addWidget(self.leTitle, 1, 1)
    leftSide.addWidget(QLabel('Description'), 2, 0)
    self.leDescription = QLineEdit(self.docProject['comment'])
    leftSide.addWidget(self.leDescription, 2, 1)
    leftSide.addWidget(QLabel('Keywords'), 3, 0)
    self.leKeywords = QLineEdit(', '.join(self.docProject['tags']))
    leftSide.addWidget(self.leKeywords, 3, 1)
    leftSide.addWidget(QLabel('Category/Community'), 4, 0)
    self.leCategory = QLineEdit(repositories.get('category',''))
    leftSide.addWidget(self.leCategory, 4, 1)
    leftSide.addWidget(QLabel('Additional'), 5, 0)
    self.leAdditional = QLineEdit(json.dumps(repositories.get('additional',{})))
    leftSide.addWidget(self.leAdditional, 5, 1)

    self.allDocTypes = [[k,v['title']] for k,v in self.comm.docTypesTitles.items()]
    projectString = ', '.join(i[1] for i in self.allDocTypes if i[0].startswith('x'))
    _, rightSide = widgetAndLayout('V', center, spacing='m', right='l')
    rightSide.setAlignment(Qt.AlignTop)                                           # type: ignore[attr-defined]
    Label('Include item types','h2', rightSide)
    self.allCheckboxes = [QCheckBox(projectString, self)]
    self.allCheckboxes[0].setChecked(True)
    self.allCheckboxes[0].setDisabled(True)
    rightSide.addWidget(self.allCheckboxes[0])
    for i in self.allDocTypes:
      if not i[0].startswith('x') and '/' not in i[0]:
        checkbox = QCheckBox(i[1], self)                                     # pylint: disable=qt-local-widget
        checkbox.setChecked(True)
        self.allCheckboxes.append(checkbox)
        rightSide.addWidget(checkbox)

    #final button box
    _, buttonLineL = widgetAndLayout('H', self.mainL, 'm')
    buttonLineL.addStretch(1)
    if 'zenodo' in repositories and 'url' in repositories['zenodo']:
      TextButton('Upload to Zenodo',    self, [Command.UPLOAD, True],  buttonLineL)
    if 'dataverse' in repositories and 'url' in repositories['dataverse']:
      TextButton('Upload to Dataverse', self, [Command.UPLOAD, False], buttonLineL)
    TextButton('Cancel',              self, [Command.CANCEL], buttonLineL, 'Discard changes')


  def execute(self, command:list[Any]) -> None:
    """
    Event if user clicks button in the center

    Args:
      command (list): list of commands
    """
    if command[0] is Command.CANCEL:
      self.reject()
    elif command[0] is Command.UPLOAD:
      # collect docTypes and create .eln
      docTypes = [i.text() for i in self.allCheckboxes if i.isChecked() and ', ' not in i.text()]
      # collect metadata and save parts of it
      metadata = {'title': self.leTitle.text(),
                  'description': self.leDescription.text(),
                  'keywords': [i.strip() for i in self.leKeywords.text().split(',')],
                  'category': self.leCategory.text(),
                  'additional': json.loads(self.leAdditional.text()),
                  'author': self.comm.configuration['authors'][0]}
      repositories = self.comm.configuration['repositories']
      repositories['category'] = self.leCategory.text()
      repositories['additional'] = json.loads(self.leAdditional.text())
      with open(Path.home()/CONF_FILE_NAME, 'w', encoding='utf-8') as fConf:
        fConf.write(json.dumps(self.comm.configuration,indent=2))
      self.comm.uiRequestTask.emit(Task.SEND_REPOSITORY, {'projID': self.comm.projectID, 'docTypes': docTypes,
                'uploadZenodo':command[1], 'repositories': repositories, 'metadata': metadata})
      self.accept()
    else:
      print('Got some button, without definition', command)
    return


  def changeButtonOnTest(self, success:bool, button:TextButton, message:str='') -> None:
    """ Helper function to change buttons upon success/failure

    Args:
      success (bool): change to successful state
      button (TextBotton): button to change
      message (str): optional message on failure
    """
    if success:
      button.setStyleSheet('background: #00FF00')
      button.setText('')
      button.setIcon(qta.icon('fa5s.check-square', scale_factor=1))
    else:
      if message:
        showMessage(self, 'Error', message)
      button.setStyleSheet('background: #FF0000')
      button.setText('')
      button.setIcon(qta.icon('fa5.times-circle', scale_factor=1))
    return


  def reject(self) -> None:
    """ Reject the dialog, stop the thread and disconnect signals """
    self.comm.backendThread.worker.beSendDoc.disconnect(self.onGetData)
    super().reject()


  def accept(self) -> None:
    """ Accept the dialog, stop the thread and disconnect signals """
    self.comm.backendThread.worker.beSendDoc.disconnect(self.onGetData)
    super().accept()


class Command(Enum):
  """ Commands used in this file """
  UPLOAD = 1
  CANCEL = 2



""" filter manager for table models in pasta_eln UI"""
from enum import Enum
from PySide6.QtCore import QSortFilterProxyModel, Qt
from PySide6.QtGui import QStandardItemModel
from PySide6.QtWidgets import QComboBox, QLineEdit, QWidget
from .guiStyle import IconButton


class FilterItem:
  """Represents a single filter with its model and associated widgets"""

  def __init__(self, filterID: int, headerOptions: list[str], parentWidget:QWidget):
    """ Initialize

    Args:
      filterID (int): Unique identifier for the filter
      headerOptions (list[str]): List of column headers to filter on
      parentWidget: Parent widget for GUI elements
    """
    self.id = filterID
    self.model = QSortFilterProxyModel()
    self.model.setFilterCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
    self.model.setFilterKeyColumn(0)

    # Create widgets
    self.select = QComboBox()
    self.select.addItems(headerOptions + ['rating'] if 'tag' in headerOptions else headerOptions)
    self.select.setMinimumWidth(max(len(i) for i in headerOptions) * 14)
    self.text = QLineEdit('')
    self.inverse = IconButton('ph.selection-inverse-fill', parentWidget,
                              [FilterCommand.SET_FILTER, filterID, 'invert'], None, checkable=True)
    self.delete = IconButton('fa5s.minus-square', parentWidget,
                              [FilterCommand.DELETE_FILTER, filterID], None)


class FilterManager:
  """Manages the chain of filters applied to a table model"""

  def __init__(self, baseModel: QStandardItemModel):
    """ Initialize the FilterManager
    Args:
      baseModel (QStandardItemModel): The initial model to apply filters to
    """
    self.baseModel = baseModel
    self.filters: dict[int, FilterItem] = {}
    self.nextID = 1
    self.modelChain: list[QStandardItemModel | QSortFilterProxyModel] = [baseModel]


  def addFilter(self, headerOptions: list[str], parentWidget:QWidget) -> FilterItem:
    """Add a new filter to the chain"""
    filterID = self.nextID
    self.nextID += 1
    filterItem = FilterItem(filterID, headerOptions, parentWidget)
    self.filters[filterID] = filterItem
    # Link to the current end of the chain
    filterItem.model.setSourceModel(self.modelChain[-1])
    self.modelChain.append(filterItem.model)
    return filterItem


  def removeFilter(self, filterID: int) -> None:
    """Remove a filter and rebuild the chain"""
    if filterID not in self.filters:
      return
    del self.filters[filterID]
    self.rebuildChain()
    return


  def rebuildChain(self) -> None:
    """Rebuild the model chain after filter removal"""
    self.modelChain = [self.baseModel]
    for filterID in sorted(self.filters.keys()):
      filterItem = self.filters[filterID]
      filterItem.model.setSourceModel(self.modelChain[-1])
      self.modelChain.append(filterItem.model)


  def getFinalModel(self) -> QStandardItemModel | QSortFilterProxyModel:
    """Get the final model in the chain (for table display)"""
    return self.modelChain[-1]


  def getFilter(self, filterID: int) -> FilterItem | None:
    """Get a specific filter by ID"""
    return self.filters.get(filterID)


  def clearAll(self) -> None:
    """Remove all filters"""
    self.filters.clear()
    self.modelChain = [self.baseModel]


  def setBaseModel(self, newBaseModel: QStandardItemModel) -> None:
    """Update the base model and rebuild chain"""
    self.baseModel = newBaseModel
    self.modelChain[0] = newBaseModel
    self.rebuildChain()


class FilterCommand(Enum):
  """ Commands used in this file """
  ADD_FILTER       = 1
  DELETE_FILTER    = 2
  SET_FILTER       = 3



""" Table Header dialog: change which columns are shown and in which order """
import logging
from collections import defaultdict
from enum import Enum
from typing import Any
import pandas as pd
from PySide6.QtCore import Slot
from PySide6.QtWidgets import QComboBox, QDialog, QDialogButtonBox, QListWidget, QVBoxLayout
from ..backendWorker.sqlite import MAIN_ORDER
from ..backendWorker.worker import Task
from ..fixedStringsJson import tableHeaderHelp
from .guiCommunicate import Communicate
from .guiStyle import IconButton, widgetAndLayout
from .messageDialog import showMessage


class TableHeader(QDialog):
  """ Table Header dialog: change which columns are shown and in which order """
  def __init__(self, comm:Communicate, docType:str):
    """
    Initialization

    Args:
      comm (Communicate): communication channel
      docType (string):  document type
    """
    super().__init__()
    self.comm    = comm
    self.docType = docType
    self.allSet  = set(MAIN_ORDER)
    self.allSet  = {i[1:] if i[0]=='.' else i for i in self.allSet}
    self.selectedList:list[str] = []
    self.otherDict:dict[str, list[str]] = {}

    # GUI elements
    self.setWindowTitle('Select list columns')
    self.setMinimumWidth(600)
    mainL = QVBoxLayout(self)
    _, bodyL = widgetAndLayout('H', mainL, spacing='m')

    _, leftL = widgetAndLayout('V', bodyL, spacing='m')
    self.choicesW = QListWidget()
    self.choicesW.setMinimumHeight(250)
    leftL.addWidget(self.choicesW)
    _, otherL = widgetAndLayout('H', leftL, spacing='s')
    self.other1 = QComboBox()
    self.other1.addItems(sorted(self.otherDict.keys()))
    self.other1.currentTextChanged.connect(self.changeCB1)
    otherL.addWidget(self.other1)
    self.other2 = QComboBox()
    self.other2.addItems(self.otherDict.get(self.other1.currentText(),[]))
    self.other2.activated.connect(self.changeCB2)
    otherL.addWidget(self.other2)

    _, centerL = widgetAndLayout('V', bodyL)
    IconButton('fa5s.angle-right',        self, [Command.ADD],       centerL, 'add to right')
    IconButton('fa5s.angle-left',         self, [Command.DELETE],    centerL, 'delete from right')
    IconButton('fa5s.angle-up',           self, [Command.MOVE_UP],   centerL, 'move up')
    IconButton('fa5s.angle-down',         self, [Command.MOVE_DOWN], centerL, 'move down')
    self.selectW = QListWidget()
    bodyL.addWidget(self.selectW)
    #final button box
    buttonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Save | QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Help)
    buttonBox.clicked.connect(self.closeDialog)
    mainL.addWidget(buttonBox)
    # request data
    self.comm.backendThread.worker.beSendSQL.connect(self.onGetData)
    self.sqlCmd1 = f'SELECT view FROM docTypes WHERE docType LIKE "{docType}%"'
    self.sqlCmd2 = f'SELECT DISTINCT properties.key FROM properties JOIN main USING(id) WHERE main.type LIKE "{docType}%"'
    self.comm.uiSendSQL.emit([{'type':'get_df', 'cmd':self.sqlCmd1}])
    self.comm.uiSendSQL.emit([{'type':'get_df', 'cmd':self.sqlCmd2}])



  @Slot(str, pd.DataFrame)
  def onGetData(self, cmd:str, data:pd.DataFrame) -> None:
    """ Callback function to handle the received data
    Args:
      cmd (str): command that was sent
      data (pd.DataFrame): DataFrame containing table
    """
    if cmd == self.sqlCmd1:
      self.selectedList = data.values[0][0].split(',')
      self.selectedList = [i[1:] if i[0]=='.' else i for i in self.selectedList]
      self.allSet       = self.allSet.union(self.selectedList)
      self.paint()
    elif cmd == self.sqlCmd2:
      otherList = [i.split('.', 1) for i in data['key'].to_list()]
      otherDict = defaultdict(set)
      for k,v in otherList:
        otherDict[k].add(v)
      self.otherDict = {k:sorted(v) for k,v in otherDict.items()}
      self.other1.clear()
      self.other2.clear()
      self.other1.addItems(sorted(self.otherDict.keys()))
      self.other2.addItems(self.otherDict.get(self.other1.currentText(),[]))


  def paint(self) -> None:
    """ Paint the dialog with the current data """
    self.choicesW.clear()
    self.selectW.clear()
    self.choicesW.addItems(list(self.allSet.difference(self.selectedList)))
    self.selectW.addItems(self.selectedList)


  def execute(self, command:list[Any]) -> None:
    """
    Event if user clicks button in the center

    Args:
      command (list): list of commands
    """
    selectedLeft   = [i.text() for i in self.choicesW.selectedItems()]
    selectedRight  = [i.text() for i in self.selectW.selectedItems()]
    oldIndex, newIndex = -1, -1
    if command[0] is Command.ADD:
      self.selectedList += selectedLeft
    elif command[0] is Command.DELETE:
      self.selectedList = [i for i in self.selectedList if i not in selectedRight or i in ['name'] ]
    elif command[0] is Command.MOVE_UP  and len(selectedRight)==1:
      oldIndex = self.selectedList.index(selectedRight[0])
      if oldIndex>0:
        newIndex = oldIndex-1
    elif command[0] is Command.MOVE_DOWN and len(selectedRight)==1:
      oldIndex = self.selectedList.index(selectedRight[0])
      if oldIndex<len(self.selectedList)-1:
        newIndex = oldIndex+1
    else:
      logging.error('Menu unknown: %s',command, exc_info=True)
    #change content
    if oldIndex>-1 and newIndex>-1:
      self.selectedList.insert(newIndex, self.selectedList.pop(oldIndex))
    self.choicesW.clear()
    self.choicesW.addItems(list(self.allSet.difference(self.selectedList)))
    self.selectW.clear()
    self.selectW.addItems(self.selectedList)
    if oldIndex>-1 and newIndex>-1:
      self.selectW.setCurrentRow(newIndex)
    return

  def changeCB1(self, text:str) -> None:
    """ change content of 2nd combobox based on 1st combobox
    Args:
      text (str): text of 1st combobox
    """
    self.other2.clear()
    self.other2.addItems(self.otherDict.get(text,[]))
    return

  def changeCB2(self, idx:int) -> None:
    """ add entry to the right list
    Args:
      idx (int): index of 2nd combobox
    """
    newEntry = f'{self.other1.currentText()}.{self.other2.itemText(idx)}'
    if newEntry not in self.selectedList:
      self.selectedList.append(newEntry)
      self.paint()
    return


  def closeDialog(self, btn:IconButton) -> None:
    """ save selectedList to configuration and exit """
    if btn.text().endswith('Cancel'):
      self.reject()
    elif btn.text().endswith('Save'):
      newList = [i if i in MAIN_ORDER+['tags','qrCodes'] or '.' in i else f'.{i}' for i in self.selectedList]
      self.comm.uiRequestTask.emit(Task.SEND_TBL_COLUMN, {'docType':self.docType, 'newList':newList})
      self.accept()
    elif btn.text().endswith('Help'):
      showMessage(self, 'Help on individual entry', tableHeaderHelp)
    else:
      print('dialogTableHeader: did not get a fitting btn ',btn.text())
    return


  def reject(self) -> None:
    """ Reject the dialog, stop the thread and disconnect signals """
    self.comm.backendThread.worker.beSendSQL.disconnect(self.onGetData)
    super().reject()


  def accept(self) -> None:
    """ Accept the dialog, stop the thread and disconnect signals """
    self.comm.backendThread.worker.beSendSQL.disconnect(self.onGetData)
    super().accept()


class Command(Enum):
  """ Commands used in this file """
  ADD       = 1
  DELETE    = 2
  MOVE_UP   = 3
  MOVE_DOWN = 4
  USE_TEXT  = 5



#!/usr/bin/python3
"""Commandline utility to admin local installation and convert from Pasta-ELN version 2"""
import json
import os
import platform
import re
import shutil
import sys
import traceback
from sqlite3 import IntegrityError
from typing import Any, Callable, Union
from pasta_eln.backendWorker.backend import Backend
from pasta_eln.backendWorker.elabFTWsync import Pasta2Elab
from pasta_eln.fixedStringsJson import defaultDocTypes, defaultSchema
from pasta_eln.miscTools import getConfiguration
from pasta_eln.textTools.stringChanges import outputString


class Tools:
  """Commandline utility to admin local installation and convert from Pasta-ELN version 2"""


  def __init__(self) -> None:
    self.backend:Backend|None = None
    self.projectGroup = ''
    self.configuration, _ = getConfiguration(self.projectGroup)


  def __choice__(self, command:str) -> str:
    """  depending on choice, execute command
    Args:
      command (str): command
    """
    helpString = 'Command line program to troubleshoot and possibly repair Pasta-ELN database\n'\
                'Commands - general:\n'\
                '  [h]elp: long help\n'\
                '  [q]uit\n'
    helpString += '  [p]rint a document\n'
    if command == 'p':
      self.printOrDelete()
    helpString += '  [d]elete a document\n'
    if command == 'd':
      self.printOrDelete(projectGroup='', docID='', output=False)
    helpString += '  [s]can all projects\n'
    if command == 's':
      self.scanAllProjects()

    helpString += 'Commands - update:\n'
    helpString += '  [u3] convert from one procedure to 3 workflows\n'
    if command == 'u3':
      self.update3Workflow()

    helpString += 'Commands - database integrity:\n'
    helpString += '  [v]erify\n'
    if command == 'v':
      self.verifyPasta()
    helpString += '  [r]epair\n'
    if command == 'r':
      self.verifyPasta('', self.__userQuestion__)
    helpString += '  [rA]epair: answer always "yes": WARNING can change too much\n'
    if command == 'rA':
      self.verifyPasta('', lambda t: self.__userQuestion__(t, True))
    helpString += '  [rS]repair schema: reset the doc-type schema to its original state\n'
    if command == 'rS':
      self.repairSchema()
    helpString += '  [rp1] repair properties: add missing "."\n'
    if command == 'rp1':
      self.repairPropertiesDot()
    helpString += '  [cp]-create a lost and found project: helpful for some repair operations\n'
    if command == 'cp':
      self.createLostAndFound()

    helpString += 'Commands - clone from / to server:\n'
    helpString += '  [ss]ync SEND ALL\n'
    helpString += '  [sg]ync GET ALL\n'
    if command in {'ss','sg'}:
      self.sync('', command)
    helpString += '  [pL]urge local database: REMOVE EVERYTHING\n'
    if command == 'pL':
      self.purgeLocal()
    helpString += '  [pR]urge remote database: REMOVE EVERYTHING\n'
    if command == 'pR':
      self.purgeRemote()

    return helpString if command=='h' else 'Done'


  def __setBackend__(self, projectGroup:str='') -> None:
    """ Internal function: return backend based on project group
    Args:
      projectGroup (str): "name" of project group
    """
    projectGroupKeys = list(self.configuration['projectGroups'].keys())
    print('Project groups:','  '.join(f'{idx+1}-{i}' for idx,i in enumerate(projectGroupKeys)))
    while not self.projectGroup:
      projectGroup = projectGroup or input('  Enter project group [number or name or entire text]: ').strip()
      if re.match(r'^\d+-\w+$', projectGroup):
        self.projectGroup = projectGroup.split('-')[1]
      elif re.match(r'^\d+$', projectGroup):
        self.projectGroup = projectGroupKeys[int(projectGroup)-1]
      elif projectGroup in projectGroupKeys:
        self.projectGroup = projectGroup
      else:
        projectGroup = ''
    self.backend = Backend(self.configuration, self.projectGroup)
    return


  def __progressBar__(self, iteration:int, total:int=100, prefix:str='', suffix:str='', decimals:int=1,
                      length:int=100, fill:str='█', printEnd:str='') -> None:
    """
    Call in a loop to create terminal progress bar
    - https://stackoverflow.com/questions/3173320/text-progress-bar-in-terminal-with-block-characters

    Args:
        iteration (int): current iteration
        total     (int): total iterations
        prefix    (str): prefix string
        suffix    (str): suffix string
        decimals  (int): positive number of decimals in percent complete
        length    (int): character length of bar
        fill      (int): bar fill character
        printEnd  (int): end character (e.g. "\r", "\r\n")
    """
    percent = ('{0:.' + str(decimals) + 'f}').format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    barString = fill * filledLength + '-' * (length - filledLength)
    print(f'\r{prefix} |{barString}| {percent}% {suffix}', end = printEnd)
    # Print New Line on Complete
    if iteration == total:
      print()
    return


  def __updateProgressBar__(self, dType:str, data:str|int) -> None:
    """ update dialog
    - "text" and "append" will update the text

    Args:
      dType (str): what to update and how "text", "append", "count", "incr"
      data (str): str- or int-value to update with
    """
    if dType in {'text','append'} and isinstance(data, str):
      print(data)
    if dType=='count' and isinstance(data, int):
      self.__progressBar__(data)
    return


  def __userQuestion__(self, errorMessage:str, allYes:bool=False) -> bool:
    """ Ask user if repair should be done

    Args:
      errorMessage (str): error message
      allYes (bool): answer all questions with yes

    Return:
      bool: repair should be done
    """
    print(errorMessage)
    return True if allYes else input('Repair [yN]: ')=='y'


  def run(self, todos:list[str]=[]) -> None:
    """  run the list of given commands, else run while-loop until user ends
    Args:
      todos (list): string-commands
    """
    if todos:
      self.__setBackend__(todos.pop(0))
    while True:
      command = todos.pop(0) if todos else input('\nImportant commands [h]elp; [q]uit; [v]erify; [r]epair: ')
      if command == 'q':
        break
      print(self.__choice__(command))
    return


  def printOrDelete(self, projectGroup:str='', docID:str='', output:bool=True) -> None:
    """ Print or delete a document
    Args:
      projectGroup (str): from which project group to delete
      docID (str): which ID to delete
      output (bool): print=True or delete=False
    """
    if not self.projectGroup:
      self.__setBackend__(projectGroup)
    if self.backend is None:
      return
    if not docID:
      docID = input('Enter docID: ').strip()
    if output:
      print(json.dumps(self.backend.db.getDoc(docID), indent=2))
    else:
      self.backend.db.remove(docID)
    return


  def scanAllProjects(self, projectGroup:str='') -> None:
    """ Scan all projects to identify missing items

      Args:
        projectGroup (str): name of project group
    """
    if not self.projectGroup:
      self.__setBackend__(projectGroup)
    if self.backend is None:
      return
    projectGroups = self.backend.db.getView('viewDocType/x0')['id'].values
    for projID in projectGroups:
      self.backend.scanProject(lambda i: self.__updateProgressBar__('count',i), projID)
    return


  def translateDoc(self, doc:dict[str,Any], comment:str='') -> tuple[dict[str,Any],list[Any]]:
    """ translate to new style

    Args:
      doc (dict): input document
      comment (str): comment to add to the output, e.g. offending doc-id

    Returns:
      dict: output document
    """
    from .textTools.handleDictionaries import fillDocBeforeCreate
    defaultValues = {'gui':[True,True], 'user':''}
    try:
      doc['id'] = doc.pop('_id')
      for key in ('name','user','type','gui','tags','client','branch','date'):
        if key in doc:                                               #skip if key is already in correct format
          continue
        if key in defaultValues and f'-{key}' not in doc:
          doc[key] = defaultValues[key]
          continue
        doc[key] = doc.pop(f'-{key}')
      doc['dateCreated']  = doc['date']
      doc['dateModified'] = doc.pop('date')
      doc['branch'] = doc['branch'][0] | {'op':'c'} if doc['branch'] else \
                      {'stack':[], 'child':9999, 'path':None, 'show':[True], 'op':'c'}
      del doc['_rev']
      attachments = doc.pop('_attachments',[])
      doc = fillDocBeforeCreate(doc, doc['type'])
    except Exception:
      print('Input document has mistakes in: ',comment,'\n',json.dumps(doc, indent=2))
      raise
    return doc, attachments


  def update3Workflow(self, projectGroup:str='') -> None:
    """ Update from a 1 procedure into a 3 workflow (plan, procedure, log) version
    - Starting at 3.1 this new style is the default

    Args:
      projectGroup (str): name of project group
    """
    if not self.projectGroup:
      self.__setBackend__(projectGroup)
    if self.backend is None:
      return
    cursor = self.backend.db.cursor

    cursor.execute("DELETE FROM docTypes WHERE docType = 'procedure'")
    for row in defaultDocTypes:
      if 'workflow' in row[0]:
        cmd = f"INSERT INTO docTypes VALUES ({', '.join(['?']*len(row))});"
        try:
          cursor.executemany(cmd, [row])
        except IntegrityError:
          print(f'Already up to date: docTypes {row[0]}')

    cursor.execute("DELETE FROM docTypeSchema WHERE docType = 'procedure'")
    for rowS in defaultSchema:
      if 'workflow' in str(rowS[0]):
        cmd = f"INSERT INTO docTypeSchema VALUES ({', '.join(['?']*len(rowS))});"
        try:
          cursor.executemany(cmd, [rowS])
        except IntegrityError:
          print(f'Already up to date: docTypeSchema {rowS[0]}')

    cursor.execute('SELECT type from main WHERE type LIKE "procedure%"')
    docTypesOld = {i[0] for i in cursor.fetchall()}
    for docType in docTypesOld:
      cmd =f'UPDATE main SET type="workflow/{docType}" WHERE type = "{docType}"'
      self.backend.db.cursor.execute(cmd)

    cursor.execute('SELECT key from properties WHERE key LIKE ".procedure%"')
    docTypesOld = {i[0] for i in cursor.fetchall()}
    for docType in docTypesOld:
      cmd =f"UPDATE properties SET key='.workflow/{docType[1:]}' WHERE key = '{docType}'"
      self.backend.db.cursor.execute(cmd)

    cursor.execute('SELECT key from definitions WHERE key LIKE "procedure%"')
    docTypesOld = {i[0] for i in cursor.fetchall()}
    for docType in docTypesOld:
      cmd =f"UPDATE definitions SET key='workflow/{docType}' WHERE key = '{docType}'"
      self.backend.db.cursor.execute(cmd)

    self.backend.db.connection.commit()
    return


  def verifyPasta(self, projectGroup:str='', repair:Union[None,Callable[[str],bool]]=None) -> None:
    """ Do the default verification of PastaELN. Adopted to CLI
      Args:
      projectGroup (str): name of project group
      repair (function): repair
    """
    if not self.projectGroup:
      self.__setBackend__(projectGroup)
    if self.backend is None:
      return
    outputString('print','info', self.backend.checkDB(outputStyle='text', repair=repair, minimal=False))
    # print hierarchy of project with name
    # projName = '...'
    # df = self.backend.db.getView('viewDocType/x0')
    # self.backend.changeHierarchy(df[df['name']==projName]['id'].values[0])
    # print(self.backend.outputHierarchy(False,True))
    return


  def repairSchema(self, projectGroup:str='') -> None:
    """ Do the default verification of PastaELN. Adopted to CLI
      Args:
      projectGroup (str): name of project group
      repair (function): repair
    """
    if not self.projectGroup:
      self.__setBackend__(projectGroup)
    if self.backend is None:
      return
    self.backend.db.dataHierarchyInit(True)
    return


  def createLostAndFound(self, projectGroup:str='') -> None:
    """ Create a lost and found project that is the target of misc documents
      Args:
        projectGroup (str): name of project group
    """
    if not self.projectGroup:
      self.__setBackend__(projectGroup)
    if self.backend is None:
      return
    self.backend.addData('x0', {'name': 'Lost and Found', 'comment': 'Location of lost database items'})
    return


  def purgeLocal(self, projectGroup:str='') -> None:
    """ Purge entire local storage, remove everything

      Args:
      projectGroup (str): name of project group
    """
    if not self.projectGroup:
      self.__setBackend__(projectGroup)
    if self.backend is None:
      return
    dirName = self.backend.basePath
    if dirName.as_posix() != '/home/steffen/FZJ/pasta_misc/testing':                 #TODO temporary safeguard
      print('DO NOT DELETE THIS')
      return
    self.backend.exit()
    try:
      shutil.rmtree(dirName)
      os.makedirs(dirName)
      if platform.system()=='Windows':
        print('Try-Except unnecessary')
    except Exception:
      pass
    self.__setBackend__(self.projectGroup)
    return


  def purgeRemote(self, projectGroup:str='') -> None:
    """ Purge entire remote storage, remove everything

      Args:
      projectGroup (str): name of project group
    """
    if not self.projectGroup:
      self.__setBackend__(projectGroup)
    if self.backend is None:
      return
    Pasta2Elab(self.backend, self.projectGroup, True)
    return



  def sync(self, projectGroup:str='', command:str='') -> None:
    """ synchronize with elab server

      Args:
      projectGroup (str): name of project group
      command (str): 'ss' or 'sg' to send/get data
    """
    if not self.projectGroup:
      self.__setBackend__(projectGroup)
    if self.backend is None:
      return
    syncObj = Pasta2Elab(self.backend, self.projectGroup)
    if command:
      syncObj.sync('sA' if command=='ss' else 'gA', progressCallback=self.__updateProgressBar__)
    return


  def repairPropertiesDot(self, projectGroup:str='') -> None:
    """ Repair sqlite database by changing properties table: if key does not have a ., prepend it
    Args:
      projectGroup (str): name of project group
    """
    if not self.projectGroup:
      self.__setBackend__(projectGroup)
    if self.backend is None:
      return
    key = input('Which key to repair, e.g. "chemistry" will become .chemistry? ')
    self.backend.db.cursor.execute(f"SELECT id FROM properties where key == '{key}'")
    res = self.backend.db.cursor.fetchall()
    for idx, docID in enumerate(res):
      docID = docID[0]
      try:
        self.backend.db.cursor.execute(f"UPDATE properties SET key='.{key}' WHERE id == '{docID}' and  key=='{key}'")
      except Exception:
        print(f"Error, could not change {docID} and {key}. Likely that combination exists already in properties. Repair manually")
        if idx==0:
          print(traceback.format_exc())
    self.backend.db.connection.commit()
    print('Done')
    return



def main() -> None:
  '''
  Main function
  '''
  print('\n-------------------------------------------------------------------------')
  print(  '------------     Manual functions for PASTA-ELN installation    ---------')
  print(  '-------------------------------------------------------------------------')
  tools = Tools()
  todos = []
  if len(sys.argv)==2:
    todos = sys.argv[1].split()
    if not re.match(r'^\d+$', todos[0]):
      print('** ERROR Malformed todo list (start with a number), exit')
      return
  tools.run(todos)
  return


if __name__ ==  '__main__':
  main()
